#include "ATLAS_SUSY_2019_09_onshell.h"


/**
* @file
* Search for chargino--neutralino pair production in final states with three leptons and missing transverse momentum
* Targets WZ signature with W->lv and Z->l+l-
* Published in 
* 2106.01676
* HEPData-ins-1866951
* https://www.hepdata.net/record/ins1866951
* https://atlas.web.cern.ch/Atlas/GROUPS/PHYSICS/PAPERS/SUSY-2019-09/
*
* This is the "onshell WZ" version
* based on pseudocode & GAMBIT version
* Recast be Mark Goodsell
**/




namespace HEP
{
    namespace ATLAS_SUSY_2019_09_onshell 
    {
/* Place here your user-defined functions */

double ATLASJER(double &x)  // x is here PT
{
    // Jet energy resolution interpolated from figure 8 of 2007.02645
    return 0.0313508 + 0.0183674/x + 0.710639/sqrt(x);
}


bool radiusCalcLepton(const HEP::Particle* lepton,const HEP::Jet* jet) { 

           double drspecial=0.04 + 10./lepton->pT();
           if (drspecial < 0.4)
           {
            if(lepton->mom().deltaR_eta(jet->mom()) > drspecial)
            {
                return false;
            }
           }
           
        return true;
	    };

         bool LessThan3Tracks(const HEP::Jet* jet,const HEP::Event* event) { 
            if(jet->btag()) return false; // keep bjets!
           double dr=0.4;
           
            int ntracks=0;

           for (auto track: event->get_charged_hadrons()) // 
           {
            if((track->pT() > 0.5) && (jet->mom().deltaR_eta(track->mom()) < dr))
            {
               ntracks++;
            }
               if (ntracks>=3) {return false;}

           }
           
           
        return true;
	    };
double diElectrontriggerEff(double &x)
      {
          // fitted function for *one leg* of double electron trigger efficiency
         double x2,x3;
         x2=x*x;
         x3=x2*x;
         
         return 0.98 - 945.54/x3 - 128.24/x2 + 1.653/x;

      };

 double doubleMuontriggerEff(const HEP::Particle *muon)  // check for eta and pT
      {
         double pT = muon->pT();
         if(pT < 14.6) return 0.0;
         if(muon->abseta() < 1.05)
         {
         
            return 0.75;

         }
         else
         {
            return 0.85;
         }
      };


    }
}

using namespace HEP::ATLAS_SUSY_2019_09_onshell;

void ATLAS_SUSY_2019_09_onshell::init() {
this->DetectorFunction="ATLAS";
cout << "---------------------------------------------------" << endl;
cout << "-- Analysis ATLAS_SUSY_2019_09_onshell --" << endl;
cout << "---------------------------------------------------" << endl;

const std::vector<std::string> allSFs={"SFOSONSSR1", "SFOSONSSR2", "SFOSONSSR3", "SFOSONSSR4", "SFOSONSSR5",
  "SFOSONSSR6", "SFOSONSSR7", "SFOSONSSR8", "SFOSONSSR9", "SFOSONSSR10",
  "SFOSONSSR11", "SFOSONSSR12", "SFOSONSSR13", "SFOSONSSR14", "SFOSONSSR15",
  "SFOSONSSR16", "SFOSONSSR17", "SFOSONSSR18", "SFOSONSSR19", "SFOSONSSR20"};


const std::vector<std::string> allDFs={"DFOSSR1", "DFOSSR2"};

for(auto region : allSFs)
{
    AddRegionSelection(region);
}

const std::vector<std::string> nojetSFs={"SFOSONSSR1", "SFOSONSSR2", "SFOSONSSR3", "SFOSONSSR4", "SFOSONSSR5",
  "SFOSONSSR6", "SFOSONSSR7", "SFOSONSSR8"};
const std::vector<std::string> jetSFs={"SFOSONSSR9", "SFOSONSSR10",
  "SFOSONSSR11", "SFOSONSSR12", "SFOSONSSR13", "SFOSONSSR14", "SFOSONSSR15",
  "SFOSONSSR16", "SFOSONSSR17", "SFOSONSSR18", "SFOSONSSR19", "SFOSONSSR20"};
const std::vector<std::string> lowHTSFs={"SFOSONSSR9", "SFOSONSSR10",
  "SFOSONSSR11", "SFOSONSSR12", "SFOSONSSR13", "SFOSONSSR14", "SFOSONSSR15",
  "SFOSONSSR16"};
  const std::vector<std::string> highHTSFs={"SFOSONSSR17", "SFOSONSSR18", "SFOSONSSR19", "SFOSONSSR20"};
const std::vector<std::string> lowmTSFs={"SFOSONSSR1", "SFOSONSSR2", "SFOSONSSR3", "SFOSONSSR4","SFOSONSSR9", "SFOSONSSR10",
  "SFOSONSSR11", "SFOSONSSR12"};
 const std::vector<std::string> highmTSFs={"SFOSONSSR5",
  "SFOSONSSR6", "SFOSONSSR7", "SFOSONSSR8","SFOSONSSR13", "SFOSONSSR14", "SFOSONSSR15",
  "SFOSONSSR16"};



AddCut("3 isolated leptons and ET",allSFs);
AddCut("nSFOS >= 1",allSFs);
AddCut("Trigger",allSFs);
AddCut("bJets == 0",allSFs);
AddCut("Resonance veto mll > 12 GeV",allSFs);
AddCut("|m3l - mZ| > 15 GeV",allSFs);
AddCut("mll in [75,105]",allSFs);
AddCut("njets == 0",nojetSFs);  // for SRs 1 to 8
AddCut("njets > 0",jetSFs);  // for SRs 9 to 20
AddCut("HT < 200",lowHTSFs);  // SRs 9 to 16
AddCut("HT > 200",highHTSFs);  // SRs 17 to 20

AddCut("mT in [100,160]",lowmTSFs); // SRs 1 to 4 and 9 to 12
AddCut("mT > 160",highmTSFs);  // SRs 5 to 8 and 13 to 16
AddCut("mT > 100",highHTSFs);  // SRs 17 to 20

for(auto region : allSFs)
{
    AddCut(region,region);
    //cout << "Adding region " << region << endl;
}


/*


3 isolated lepton and ${E}_{\text{T}}$ selection , 227 
$n_{SFOS} \geq 1$ , 226 
Trigger selection ,222 
$n_{b-jets} = 0$ ,  209 
Resonance veto $m_{\ell\ell} > 12$ GeV , 209 
$|m_{3\ell}-m_{Z}| > 15 $ GeV , 203 
"$m_{\ell\ell} \in [75,105]$ GeV ", 196  
with MC to data weight , 186  
$n_{jets}$ = 0 , 76.4 
with MC to data weight , 73.3 
"$m_{T} \in [100,160]$ GeV ", 26.7 
SR$^{WZ}$-1 ,  20.9 
SR$^{WZ}$-2 ,  4.86 
SR$^{WZ}$-3 ,  0.78 
SR$^{WZ}$-4 ,  0.14 
$m_{T} > 160$ GeV , 5.80 
SR$^{WZ}$-5 ,  4.64 
SR$^{WZ}$-6 ,  0.16 
SR$^{WZ}$-7 ,  0 
SR$^{WZ}$-8 ,  0 
SR$^{WZ}_{0j}$-1 to 8 ,  31.4 
"$n_{jets} > 0, H_{T} < 200$ GeV ", 97.5 
with MC to data weight , 91.8 
"$m_{T} \in [100,160]$ GeV ", 29.6 
SR$^{WZ}$-9 ,  8.75 
SR$^{WZ}$-10 ,  3.46 
SR$^{WZ}$-11 ,  0.54 
SR$^{WZ}$-12 ,  0 
$m_{T} > 160$ GeV , 9.50 
SR$^{WZ}$-13 ,  7.19 
SR$^{WZ}$-14 ,  1.53 
SR$^{WZ}$-15 ,  0.09 
SR$^{WZ}$-16 ,  0 
$n_{jets} > 0 H_{T} > 200$ GeV , 22.2 
$H_{T}^{lep} < 350$ GeV , 20.9 
with MC to data weight , 19.3 
$m_{T} > 100$ GeV , 10.8 
SR$^{WZ}$-17 ,  2.53 
SR$^{WZ}$-18 ,  3.12 
SR$^{WZ}$-19 ,  1.09 
SR$^{WZ}$-20 ,  1.13 
SR$^{WZ}_{nj}$-9 to 20 ,  29.4 


*/




}

ATLAS_SUSY_2019_09_onshell::~ATLAS_SUSY_2019_09_onshell() {


};

void ATLAS_SUSY_2019_09_onshell::Execute(std::mt19937 &engine) {

//cout << "Execute event" <<endl;
static std::uniform_real_distribution<double> rd(0.0,1.0);
vector<const HEP::Jet*> baseJets,signalJets,bJets,jets,jets20; 
  vector<const HEP::Particle*>baseMuons,baseElectrons,baseLeptons,signalMuons,signalElectrons,signalLeptons; 



/*

 auto baselineElectrons  = event->getElectrons(4.5, 2.47, ELooseBLLH);
   auto baselineMuons      = event->getMuons(3.0, 2.5, MuMedium | MuNotCosmic | MuQoPSignificance);
   auto baselineJets       = event->getJets(20., 4.5);
   auto metVec             = event->getMET();
   double met              = metVec.Et();
   double metSig           = event->getMETSignificance();

   //====================================================================================================
   auto baselineleptons = baselineElectrons + baselineMuons;
   int nbaselineleptons = baselineleptons.size();
    */

   auto metVec             = Event->missingmom();


   // Smearing: 8.9 GeV uncertainty on MET from soft term

   static std::normal_distribution<> smearmet(0.0, 8.9);
   
   metVec.setPx(metVec.px()+smearmet(engine));
   metVec.setPy(metVec.py()+smearmet(engine));
   metVec.setPz(0.0);


   double met              = metVec.pT();

   //FillYodaHisto1D("MET",met);

/*
https://cds.cern.ch/record/2630948/files/ATLAS-CONF-2018-038.pdf page 9:

The Emiss signiﬁcance is deﬁned to test the hypothesis that the total transverse momentum carried by
invisible particles (named pinvT in the following) is equal to zero against the hypothesis that pinvT 
is diﬀerent from zero.


Looking at figure 7 from that paper, even for Z -> ee nu nu only a modest proportion of events have a
significance < 3 (the largest value required for the signal in this search) since the distribution is quite
flat.

Comparing figures 7(a) and 7(c) it is reasonable to conclude that the significance is a function of Et 

*/


   double metSig = 10.0;  // fake value since we are only injecting signal

   // Values from comparing figs 7a and 7c, we only care about the thresholds 1.5 and 3.0 for the significance.
   if(met < 35.0) metSig= 2.9;
   if(met < 15.0) metSig=1.4; 


// NB a lepton trigger is used for the SRlows; for SRhigh an ETmiss trigger is used too.
// But this means we shouldn't just use the MET trigger efficiency.

   /*
    double eff;
    if(met < 120.0) 
    {
      eff=0.0;
    } 
    else if (met < 280.0)
    {
      eff = (met-120.0)/160.0;
    }
    else
    {
      eff=1.0;
    }

    if(rd(engine) > eff) metSig=0.0;
  */

   double uncertV[2][2] ={{0.0,0.0},{0.0,0.0}};





  for(auto CurrentJet : Event->jets())
    {
      if ( CurrentJet->pT() > 20.0 && abs(CurrentJet->eta())<4.5) 
      {
         double pT = CurrentJet->pT();
         double cosphi = CurrentJet->mom().px()/pT;
         double sinphi=CurrentJet->mom().py()/pT;
         double sigmapT=pT*(0.02998661080448855 - 21.064427751806225/(pT*pT) + 4.770352126699648/pT);
         double sigmaperp=0.2*sigmapT;  // don't have good figures on this, just table 1 

         uncertV[0][0]+=sigmapT*sigmapT*cosphi*cosphi +sigmaperp*sigmaperp*sinphi*sinphi;
         uncertV[0][1]+=(sigmapT*sigmapT-sigmaperp*sigmaperp)*cosphi*sinphi;
         //uncertV[1][0]=uncertV[0][1];
         uncertV[1][1]+=sigmapT*sigmapT*sinphi*sinphi +sigmaperp*sigmaperp*cosphi*cosphi;
        //if(CurrentJet->btag() && (abs(CurrentJet->eta()) < 2.5) && (rd(engine) < 0.85))
       // {
       //   bJets.push_back(CurrentJet);
       // }
        //else
        //{
        baseJets.push_back(CurrentJet); 
        //}
      }
    }

    for(auto CurrentTrack : Event->electrons())
    {
      double pT = CurrentTrack->pT(); 
      double abseta = CurrentTrack->abseta(); 
      
      if(pT > 4.5 && abseta < 2.47)
	{
      double cosphi = CurrentTrack->mom().px()/pT;
         double sinphi=CurrentTrack->mom().py()/pT;
         double sigmapT=pT*(0.00382643 - 0.13687/(pT*pT) + 0.975747/pT);
         double sigmaperp=0.2*sigmapT; // no good figures, just table 1

         uncertV[0][0]+=sigmapT*sigmapT*cosphi*cosphi +sigmaperp*sigmaperp*sinphi*sinphi;
         uncertV[0][1]+=(sigmapT*sigmapT-sigmaperp*sigmaperp)*cosphi*sinphi;
         //uncertV[1][0]=uncertV[0][1];
         uncertV[1][1]+=sigmapT*sigmapT*sinphi*sinphi +sigmaperp*sigmaperp*cosphi*cosphi;

	  //double ptratio = sumpTisolation(CurrentTrack->mom(),Event,0.4);
	  //if (ptratio < 0.2*pt) 

       if(pT > 10) baseElectrons.push_back(CurrentTrack); 
         //if((CurrentTrack->pT() > 10.0) || rd(engine) < 0.6) baseElectrons.push_back(CurrentTrack); 
      } 

    }

  for(auto  CurrentTrack : Event->muons())
    {
      double pT = CurrentTrack->pT(); 
      double abseta = CurrentTrack->abseta(); 
      if(pT > 3. && abseta < 2.5){

         double cosphi = CurrentTrack->mom().px()/pT;
         double sinphi=CurrentTrack->mom().py()/pT;
         double sigmapT=pT*(0.0221905 + 0.000101559*pT + (1.64276e-8)*pT*pT);
         double sigmaperp=0.05*sigmapT; // no good figures, just table 1

         uncertV[0][0]+=sigmapT*sigmapT*cosphi*cosphi +sigmaperp*sigmaperp*sinphi*sinphi;
         uncertV[0][1]+=(sigmapT*sigmapT-sigmaperp*sigmaperp)*cosphi*sinphi;
         //uncertV[1][0]=uncertV[0][1];
         uncertV[1][1]+=sigmapT*sigmapT*sinphi*sinphi +sigmaperp*sigmaperp*cosphi*cosphi;
//	double ptratio = sumpTisolation(CurrentTrack->mom(),Event,0.4);
	//if (ptratio < 0.2*pt) 
  if(pT > 10.)  baseMuons.push_back(CurrentTrack);
      // if((CurrentTrack->pT() > 10.0) || rd(engine) < 0.6) baseMuons.push_back(CurrentTrack); 
	
      } 

    }

   // add soft term:

   uncertV[0][0]+=8.9*8.9;
   uncertV[1][1]+=8.9*8.9;
   uncertV[1][0]=uncertV[0][1];

   /*
      NEW: construct MET significance from "hard" electrons, muons, jets and "soft term" -- which is just an 
      uncertainty of 8.9 GeV in each direction

      baseline objects before removal are fine at this point
   
      We have the resolution measurement of pT and 

   */

   double uncertVdet=uncertV[0][0]*uncertV[1][1]-uncertV[0][1]*uncertV[1][0];
   //if(uncertVdet ==0)
   //{
  //    metSig=0.0;
   //}
   if(uncertVdet != 0.0) // otherwise use default extrapolation
   {
      metSig=sqrt((uncertV[1][1]*metVec.px()*metVec.px()+uncertV[0][0]*metVec.py()*metVec.py()-2.0*uncertV[0][1]*metVec.px()*metVec.py())/uncertVdet);

   }

   //cout << "MET: " << met << ", metSig: " << metSig << endl;

   //FillYodaHisto1D("METSig",metSig);




//cout << "MET: " << met << ", nleptons: " << (baseMuons.size()+baseElectrons.size()) << endl;

//cout << "Found " << bJets.size() << " baseline b jets " <<endl;
    //baseLeptons=baseElectrons+baseMuons;


   


 // cout << "Before first removal " << endl;

  baseElectrons  	= Removal(baseElectrons, baseMuons, 0.01);
  baseJets = Removal(baseJets,baseElectrons,0.2);
//baseJets = Removal(baseJets,baseMuons,0.2); // really need some sort of 'less than three tracks'

  vector<const HEP::Jet*> newbaseJets;
  for(auto bajet : baseJets)
  {
   bool keepit=true;
   for(auto mu : baseMuons)
   {
      // paper says 0.4, pseudocode says 0.2
      if( (mu->mom().deltaR_eta(bajet->mom()) < 0.2) &&  LessThan3Tracks(bajet,Event))
      {
         keepit=false;
         break;
      }

   }
   if(keepit) newbaseJets.push_back(bajet);
  }

  baseJets=newbaseJets;



baseElectrons= Removal(baseElectrons,baseJets,0.4,&radiusCalcLepton);
baseMuons= Removal(baseMuons,baseJets,0.4,&radiusCalcLepton);




/*

Baseline electrons are required to have
𝑝 T > 4.5 GeV and fall within the acceptance of the ID (|𝜂| < 2.47). They are further required to satisfy the
calorimeter- and tracking-based ‘Loose and B-layer likelihood’ identification [145]. Baseline muons must
have 𝑝 T > 3 GeV and |𝜂| < 2.5, and satisfy Medium identification criteria

*/


std::sort(baseElectrons.begin(),baseElectrons.end(),HEP::compare_pT);
std::sort(baseMuons.begin(),baseMuons.end(),HEP::compare_pT);
filterEfficiency(baseJets,&ATLASMediumJetEff,engine);

for (auto CurrentJet : baseJets)
{
    double abseta=fabs(CurrentJet->eta());
    if(CurrentJet->pT() > 30.0 && (abseta < 2.8))
    {
    if(CurrentJet->btag() && ( abseta< 2.5) && (rd(engine) < 0.85))
       {
         bJets.push_back(CurrentJet);
        }
        else
        {
            if(!CurrentJet->btag() && ( abseta< 2.5) && (rd(engine) < 0.04)) // miss-tag rate
            {
                bJets.push_back(CurrentJet);
            }
            else
            {
                signalJets.push_back(CurrentJet);
            }
        }
    }
}


//signalElectrons=baseElectrons; // apply efficiencies for these later
//signalMuons=baseMuons; 

/*
 if(baseElectrons.size() > 0)  FillYodaHisto1D("pTe1",baseElectrons[0]->pT());
      if( (baseElectrons.size() > 1)) FillYodaHisto1D("pTe2",baseElectrons[1]->pT());
      if( (baseElectrons.size() > 2)) FillYodaHisto1D("pTe3",baseElectrons[2]->pT());
      if((baseMuons.size() > 0)) FillYodaHisto1D("pTm1",baseMuons[0]->pT());
      if((baseMuons.size() > 1)) FillYodaHisto1D("pTm2",baseMuons[1]->pT());
      if((baseMuons.size() > 2)) FillYodaHisto1D("pTm3",baseMuons[2]->pT());

if(baseJets.size() > 0) FillYodaHisto1D("JetpT",baseJets[0]->pT());
*/



 Gambit::ColliderBit::ATLAS::applyMuonIDEfficiency2020(baseMuons, "Medium",engine);
   // Apply cuts to get signal electrons
          // Apply electron ID efficiency from "Medium" criteria in 1908.00005
          Gambit::ColliderBit::ATLAS::applyElectronIDEfficiency2020(baseElectrons, "Loose",engine);
          // Apply electron isolation efficiency from "Tight" criteria in 1908.00005


          signalElectrons=baseElectrons;
signalMuons=baseMuons;
          Gambit::ColliderBit::ATLAS::applyElectronIsolationEfficiency2020(signalElectrons, "Tight",engine);
          // Apply muon isolation efficiency from "Tight" criteria in 2012.00578
          Gambit::ColliderBit::ATLAS::applyMuonIsolationEfficiency2020(signalMuons, "Tight",engine);



baseLeptons=baseElectrons;
baseLeptons.insert( baseLeptons.end(), baseMuons.begin(), baseMuons.end() );

std::sort(baseLeptons.begin(),baseLeptons.end(),HEP::compare_pT);



int nbaselineleptons = baseLeptons.size();



signalLeptons=signalElectrons;
signalLeptons.insert( signalLeptons.end(), signalMuons.begin(), signalMuons.end() );

std::sort(signalLeptons.begin(),signalLeptons.end(),HEP::compare_pT);

int nLeptons = signalLeptons.size();
//cout << "Apply first cut" << endl;
ApplyCut((nLeptons==3) && (met > 50),"3 isolated leptons and ET");
//cout << "Did it" << endl;
if(nLeptons != 3 || met <=50) return;

 int nElectrons = signalElectrons.size();
  int nMuons     = signalMuons.size();
  int nJets      = signalJets.size();
  int nBjets     = bJets.size();


  
  // calculate variables for SR definitions
  int hasSFOS = 0.;
  double mll3lZ = 9999999.;
  double mll3ldZ = 9999999.;
  double mTWZ = 0.;
  float Zmass = 91.188;
    double HTlep=0.0;

    bool NoResonanceVeto=true;
  for ( signed int i = 0; i < nLeptons; i++ ) {
    HTlep+=signalLeptons[i]->pT();
    //lepmom+=signalLeptons[i]->mom();
    for ( signed int j = 0; j < nLeptons; j++ ) {
      for ( signed int k = 0; k < nLeptons; k++ ) {
        if ( i == j || j == k || i == k ) {
          continue;
        }
        if ( (signalLeptons[i]->get_3Q() != signalLeptons[j]->get_3Q()) &&
             (signalLeptons[i]->abspid() == signalLeptons[j]->abspid()) ) {
          hasSFOS = 1;
          float tmp_mll = ( signalLeptons[i]->mom() + signalLeptons[j]->mom() ).M();
          if (tmp_mll < 12.0) NoResonanceVeto=false;
          if ( fabs(tmp_mll - Zmass) < mll3ldZ ) {
            mll3ldZ = fabs(tmp_mll - Zmass);
            mll3lZ = tmp_mll;
            mTWZ = mTcalc_simple(signalLeptons[k]->mom(), metVec);
          }
        }
      }
    }
  }
ApplyCut(hasSFOS>0,"nSFOS >= 1");

 // double HT = sumObjectsPt(signalJets);
    
    
    //double HT=HEP::HT(signalJets);
    // Try to use JER to smear HT

    double HT=0.0;
    for (auto jet : signalJets)
    {
        double jpT = jet->pT();
        double jer = ATLASJER(jpT);
        std::normal_distribution<> smearjetpt(jpT, jer*jpT); // resolution is sigma(pT)/pT
        HT+=smearjetpt(engine); 


    }


  // computed above
  //double HTlep = sumObjectsPt(signalLeptons);
  
  double mlll = (signalLeptons[0]->mom()+signalLeptons[1]->mom()+signalLeptons[2]->mom()).M();

  bool passDFOS = false;
  double dRnear = 0;
  for (signed int i = 0; i < nLeptons; i++) {
    for (signed int j = 0; j < nLeptons; j++) {
      if ((signalLeptons[i]->get_3Q() == signalLeptons[j]->get_3Q()) &&
          (signalLeptons[i]->abspid() == signalLeptons[j]->abspid()) && (i != j)) {
        for (signed int k = 0; k < nLeptons; k++) {
          if (k != i && k != j &&
              (signalLeptons[k]->get_3Q() != signalLeptons[j]->get_3Q()) &&
              (signalLeptons[k]->abspid() != signalLeptons[j]->abspid())) {
            passDFOS = true;

            if (fabs(signalLeptons[k]->mom().deltaPhi(signalLeptons[i]->mom())) <= 
                fabs(signalLeptons[k]->mom().deltaPhi(signalLeptons[j]->mom()))) {
              dRnear = signalLeptons[i]->mom().deltaR_eta(signalLeptons[k]->mom());
            } else {
              dRnear = signalLeptons[j]->mom().deltaR_eta(signalLeptons[k]->mom());
            }
          }
        }
      }
    }
  }


// here we *only* use dilepton triggers 
bool trigger=false;

  if( (baseElectrons.size() > 1) && (baseElectrons[1]->pT() > 18.0))  // dielectron trigger
      {
         double teff=0.0;
         //double teff1,teff2;
         double pT1= baseElectrons[0]->pT();
         double pT2= baseElectrons[1]->pT();
         teff=diElectrontriggerEff(pT1)*diElectrontriggerEff(pT2);

         if(rd(engine) < teff) trigger=true;

      }

      if(!trigger && (baseElectrons.size() >2) && (baseElectrons[2]->pT() > 18.0))  // dielectron trigger, another bite at the cherry
      {
          double teff=0.0;
         //double teff1,teff2;
         double pT1= baseElectrons[0]->pT();
         double pT2= baseElectrons[2]->pT();
         teff=diElectrontriggerEff(pT1)*diElectrontriggerEff(pT2);

         if(rd(engine) < teff) trigger=true;

      }

      // now do muon triggers
      /*
       if(!trigger && (baseMuons.size() > 0) && (baseMuons[0]->pT() > 27.3)) // single muon trigger
      {
         double teff=singleMuontriggerEff(baseMuons[0]);

         if(rd(engine) < teff) trigger = true;
      }
    */

      if(!trigger && (baseMuons.size() > 1) && (baseMuons[1]->pT() > 14.7)) // double muon trigger
      {
         // second bite at the cherry for the single muon trigger
         //double teffsingle=singleMuontriggerEff(baseMuons[1]);

       //////  if(rd(engine) < teffsingle)
       ///  { 
       //     trigger = true;
      //   }
     //    else  // actually check dimuon trigger
     //    {
            double teffdouble=doubleMuontriggerEff(baseMuons[0])*doubleMuontriggerEff(baseMuons[1]);
            if(rd(engine) < teffdouble)  trigger = true;
     //    }

      }

         if(!trigger && (baseMuons.size() > 2) && (baseMuons[2]->pT() > 14.7)) // double muon trigger, second chance
      {
         // second bite at the cherry for the single muon trigger
         //double teffsingle=singleMuontriggerEff(baseMuons[1]);

       //////  if(rd(engine) < teffsingle)
       ///  { 
       //     trigger = true;
      //   }
     //    else  // actually check dimuon trigger
     //    {
            double teffdouble=doubleMuontriggerEff(baseMuons[0])*doubleMuontriggerEff(baseMuons[2]);
            if(rd(engine) < teffdouble)  trigger = true;
     //    }

      }

      /*
      if(!trigger && (baseMuons.size() > 2) && (baseMuons[2]->pT() > 6.5)) // triple muon trigger
      {
         double teffsingle=singleMuontriggerEff(baseMuons[2]);

         if(rd(engine) < teffsingle)
         { 
            trigger = true; 
         }
         
         double teffdouble=doubleMuontriggerEff(baseMuons[0])*doubleMuontriggerEff(baseMuons[2]);
         if(rd(engine) < teffdouble)  trigger = true;



         if(!trigger)
         {
            double tefftriple=tripleMuontriggerEff(baseMuons[0])*tripleMuontriggerEff(baseMuons[1])*tripleMuontriggerEff(baseMuons[2]);
            if(rd(engine) < tefftriple)  trigger = true;
         }
      }
      */

if(signalLeptons[0]->pT() < 25 || signalLeptons[1]->pT() < 20 || signalLeptons[2]->pT() < 10 ) trigger = false;
ApplyCut(trigger,"Trigger");
if(!trigger) return;

ApplyCut(nBjets==0,"bJets == 0");
if(nBjets>0) return;
ApplyCut(NoResonanceVeto,"Resonance veto mll > 12 GeV");

ApplyCut(fabs(mlll - Zmass) > 15,"|m3l - mZ| > 15 GeV");
ApplyCut(mll3lZ > 75 && mll3lZ < 105,"mll in [75,105]");

ApplyCut(nJets==0,"njets == 0");
ApplyCut(nJets > 0,"njets > 0");
ApplyCut(HT<200,"HT < 200");
ApplyCut(HT>200,"HT > 200");
ApplyCut(mTWZ > 100,"mT > 100");
ApplyCut(mTWZ>160,"mT > 160");
ApplyCut(mTWZ>100 && mTWZ < 160,"mT in [100,160]");

  //====================================================================================================
  // Regions

  bool SFOSONSSR1(false), SFOSONSSR2(false), SFOSONSSR3(false), SFOSONSSR4(false), SFOSONSSR5(false),
    SFOSONSSR6(false), SFOSONSSR7(false), SFOSONSSR8(false), SFOSONSSR9(false), SFOSONSSR10(false),
    SFOSONSSR11(false), SFOSONSSR12(false), SFOSONSSR13(false), SFOSONSSR14(false), SFOSONSSR15(false),
    SFOSONSSR16(false), SFOSONSSR17(false), SFOSONSSR18(false), SFOSONSSR19(false), SFOSONSSR20(false);

  bool DFOSSR1(false), DFOSSR2(false);

 

  bool presel_DFOS = true;
  // DFOS SRs
  if   ( passDFOS == !1 ) presel_DFOS = false;
  //if (presel_DFOS) fill("cutflow_DFOS", cutflowcounter++);
  if (  metSig<8 ) presel_DFOS = false;
 // if (presel_DFOS) fill("cutflow_DFOS", cutflowcounter++);



  if ( passDFOS == 1 && signalLeptons[2]->pT()>15 && metSig>8 ) {
    if ( nJets==0 && dRnear<1.2                                            ) { DFOSSR1 = true; }//accept("DFOSSR1"); }
    if ( (nJets==1 || nJets==2) && dRnear<1.0 && signalLeptons[2]->pT()>20  ) { DFOSSR2 = true; } //accept("DFOSSR2"); }
  }

  bool presel_SFOS = true;

  if   ( hasSFOS == !1 ) presel_SFOS = false;
  //if (presel_SFOS) fill("cutflow", cutflowcounter++);
  if   (  fabs(mlll-Zmass)<15 ) presel_SFOS = false;
  //if (presel_SFOS) fill("cutflow", cutflowcounter++);

  /// Nominal SFOS SRs and VRs
  if (hasSFOS == 1 && fabs(mlll-Zmass)>15 ){

 
    // Onshell
    if ( mll3lZ>75 && mll3lZ<105   && nJets==0 &&  mTWZ >100 && mTWZ <160 && met >50  && met <100   ) { SFOSONSSR1 = true; }//accept("SFOSONSSR1"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets==0 &&  mTWZ >100 && mTWZ <160 && met >100 && met <150   ) { SFOSONSSR2 = true; }//accept("SFOSONSSR2"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets==0 &&  mTWZ >100 && mTWZ <160 && met >150 && met <200   ) { SFOSONSSR3 = true; }//accept("SFOSONSSR3"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets==0 &&  mTWZ >100 && mTWZ <160 && met >200               ) { SFOSONSSR4 = true; }//accept("SFOSONSSR4"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets==0 &&  mTWZ >160              && met >50 && met <150    ) { SFOSONSSR5 = true; }//accept("SFOSONSSR5"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets==0 &&  mTWZ >160              && met >150 && met <200   ) { SFOSONSSR6 = true; }//accept("SFOSONSSR6"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets==0 &&  mTWZ >160              && met >200 && met <350   ) { SFOSONSSR7 = true; }//accept("SFOSONSSR7"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets==0 &&  mTWZ >160              && met >350               ) { SFOSONSSR8 = true; }//}//accept("SFOSONSSR8"); }

    if ( mll3lZ>75 && mll3lZ<105   && nJets>0  && HT <200 &&  mTWZ >100 && mTWZ <160 && met >100 && met <150   ) { SFOSONSSR9 = true; }//accept("SFOSONSSR9"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets>0  && HT <200 &&  mTWZ >100 && mTWZ <160 && met >150 && met <250   ) { SFOSONSSR10 = true;}// accept("SFOSONSSR10"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets>0  && HT <200 &&  mTWZ >100 && mTWZ <160 && met >250 && met <300   ) { SFOSONSSR11 = true;}// accept("SFOSONSSR11"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets>0  && HT <200 &&  mTWZ >100 && mTWZ <160 && met >300               ) { SFOSONSSR12 = true; }//accept("SFOSONSSR12"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets>0  && HT <200 &&  mTWZ >160              && met >50 && met <150    ) { SFOSONSSR13 = true; }//accept("SFOSONSSR13"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets>0  && HT <200 &&  mTWZ >160              && met >150 && met <250   ) { SFOSONSSR14 = true; }//accept("SFOSONSSR14"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets>0  && HT <200 &&  mTWZ >160              && met >250 && met <400   ) { SFOSONSSR15 = true;}// accept("SFOSONSSR15"); }
    if ( mll3lZ>75 && mll3lZ<105   && nJets>0  && HT <200 &&  mTWZ >160              && met >400               ) { SFOSONSSR16 = true; }//accept("SFOSONSSR16"); }

    if ( mll3lZ>75 && mll3lZ<105  && nJets>0  && HT >200 &&  HTlep<350  &&  mTWZ >100    && met >150 && met <200   ) { SFOSONSSR17 = true; }//accept("SFOSONSSR17"); }
    if ( mll3lZ>75 && mll3lZ<105  && nJets>0  && HT >200 &&  HTlep<350  &&  mTWZ >100    && met >200 && met <300   ) { SFOSONSSR18 = true;}// accept("SFOSONSSR18"); }
    if ( mll3lZ>75 && mll3lZ<105  && nJets>0  && HT >200 &&  HTlep<350  &&  mTWZ >100    && met >300 && met <400   ) { SFOSONSSR19 = true;}// accept("SFOSONSSR19"); }
    if ( mll3lZ>75 && mll3lZ<105  && nJets>0  && HT >200 &&  HTlep<350  &&  mTWZ >100    && met >400               ) { SFOSONSSR20 = true;}// accept("SFOSONSSR20"); }

    ApplyCut(SFOSONSSR1,"SFOSONSSR1");
    ApplyCut(SFOSONSSR2,"SFOSONSSR2");
    ApplyCut(SFOSONSSR3,"SFOSONSSR3");
    ApplyCut(SFOSONSSR4,"SFOSONSSR4");
    ApplyCut(SFOSONSSR5,"SFOSONSSR5");
    ApplyCut(SFOSONSSR6,"SFOSONSSR6");
    ApplyCut(SFOSONSSR7,"SFOSONSSR7");
    ApplyCut(SFOSONSSR8,"SFOSONSSR8");
    ApplyCut(SFOSONSSR9,"SFOSONSSR9");
    ApplyCut(SFOSONSSR10,"SFOSONSSR10");
    ApplyCut(SFOSONSSR11,"SFOSONSSR11");
    ApplyCut(SFOSONSSR12,"SFOSONSSR12");
    ApplyCut(SFOSONSSR13,"SFOSONSSR13");
    ApplyCut(SFOSONSSR14,"SFOSONSSR14");
    ApplyCut(SFOSONSSR15,"SFOSONSSR15");
    ApplyCut(SFOSONSSR16,"SFOSONSSR16");
    ApplyCut(SFOSONSSR17,"SFOSONSSR17");
    ApplyCut(SFOSONSSR18,"SFOSONSSR18");
    ApplyCut(SFOSONSSR19,"SFOSONSSR19");
    ApplyCut(SFOSONSSR20,"SFOSONSSR20");


  }


  //====================================================================================================
  // inclusive SR/VR

  /*
  if ( SFOSONSSR1 || SFOSONSSR2 || SFOSONSSR3 || SFOSONSSR4 || SFOSONSSR5 ||
       SFOSONSSR6 || SFOSONSSR7 || SFOSONSSR8 || SFOSONSSR9 || SFOSONSSR10 ||
       SFOSONSSR11 || SFOSONSSR12 || SFOSONSSR13 || SFOSONSSR14 || SFOSONSSR15 ||
       SFOSONSSR16 || SFOSONSSR17 || SFOSONSSR18 || SFOSONSSR19 || SFOSONSSR20 ) { SFOSONSSR=true; accept("SFOSONSSR"); }
*/


//  if ( DFOSSR1 || DFOSSR2 ) accept("DFOSSR");
//  if ( passDFOS == 1 && signalLeptons[2].Pt()>15 ) accept("DFOSSR_loose");


}

