#include "CMS_EXO_20_004.h"

/**
* @file 
* Recast of CMS monojet analysis, 137 fb^{-1} @ 13 TeV
* Starting from the MadAnalysis version
*
* https://cms-results.web.cern.ch/cms-results/public-results/publications/EXO-20-004/
* https://arxiv.org/2107.13021
*
*
**/

namespace HEP
{
    namespace CMS_EXO_20_004 
    {
/* Place here your user-defined functions */

bool electronIsolation(HEP::Particle* part, const HEP::Event* event,std::mt19937& engine)
{
  double eti=sumETIsolation(part->mom(), event, 0.5);

  if(part->pT()*0.12 < eti) return false;

  return true;

}

bool muonIsolation(HEP::Particle* part, const HEP::Event* event,std::mt19937& engine)
{
  double eti=sumETIsolation(part->mom(), event, 0.5);

  if(part->pT()*0.25 < eti) return false;

  return true;

}


double muonRecoEff(HEP::Particle* lepton)
{
  double pT, abseta;

  abseta=lepton->abseta();
  if (abseta > 2.4) return 0.0;
  pT=lepton->pT();
   if(pT < 10.0) return 0.0;

  return 0.95;


}

double electronRecoEff(HEP::Particle* lepton)
{
  double pT, abseta;

  abseta=lepton->abseta();
  if (abseta > 2.5) return 0.0;
  pT=lepton->pT();
   if(pT < 10.0) return 0.0;

  if(abseta < 1.5) return 0.95;
  return 0.85;
  

}


// the signal templates imply that the upper bin is at 1400 whereas the CMS MA5 puts 99999.
vector<int> monojet_bins = { 
  250,  280,  310,  340,  370,  400,  430,  470,  510, 550,  590,  640,  690,  740,  790,  840,  900,  960, 1020, 1090, 1160, 1250, 99999, 
}; 


double deepcsv_btag_medium(double &pt)
{
    double pt2,pt3,pt4,pt5;
    pt2 = pt*pt;
    pt3=pt*pt2;
    pt4=pt*pt3;
    pt5=pt*pt4;
    if(pt < 20.0) return 0.0;
    if(pt < 50.0) return 0.194+0.0211*pt-0.000348*pt2+2.761e-6*pt3-1.044e-8*pt4+1.499e-11*pt5;
    if(pt < 250.0) return 0.557+0.003417*pt-3.26e-5*pt2+1.506e-7*pt3-3.63e-10*pt4+3.522e-13*pt5;
    if(pt < 1000.0) return 0.768-0.00055*pt+2.876e-7*pt2;
    return 0.5;  // value at 1000

}

double deepcsv_mistag_medium(double &pt)
{
double pt2,pt3,pt4,pt5;
    pt2 = pt*pt;
    pt3=pt*pt2;
    pt4=pt*pt3;
    pt5=pt*pt4;

    if (pt < 20.0) return 0.0;
    if (pt < 225) return 0.019 - 0.00031*pt +3.39e-6*pt2-1.47e-8*pt3+2.92e-11*pt4-2.12e-14*pt5;
    if (pt < 1000.0) return 0.00328 +5.7e-5*pt+4.7e-9*pt2;
    return 0.06;

}


bool has_b_tag(vector<HEP::Jet*> jets, double ptmin, double etamax, std::mt19937 &engine){ 
  // Returns true if a b jet is found 
  static  std::uniform_real_distribution<> btag_rand;
    /*
        and are identi-
ﬁed to have originated from the hadronization of a bottom quark (“b-tagged jets”) using the
DEEPCSV algorithm with a “medium” working point, corresponding to correctly identifying
a b jet with a probability of 80% and misidentifying a light-ﬂavor quark or gluon jet with a
probability of 10% [70].

Surely 1%??? 
See figure 16 and table 2 in https://arxiv.org/pdf/1712.07158.pdf
For the medium working point we see on average epsilon_b of 68%, epsilon_c of 12% and mistag of 1.1%

The CMS analysis uses the parametrised efficiency formula from MA5
 
In fact the paper [70] provides a formula in the appendix on page 97

btag:

pt < 50:
0.194+0.0211*pt-0.000348*pt2+2.761e-6*pt3-1.044e-8*pt4+1.499e-11*pt5;
pt < 250:
0.557+0.003417*pt-3.26e-5*pt2+1.506e-7*pt3-3.63e-10*pt4+3.522e-13*pt5;
pt< 1000:
0.768-0.00055*pt+2.876e-7*pt2;

mistag:
pt < 225:
0.019 - 0.00031*pt +3.39e-6*pt2-1.47e-8*pt3+2.92e-11*pt4-2.12e-14*pt5;

225 < pt < 1000:
0.00328 +5.7e-5*pt+4.7e-9*pt2;

This has **much** better performance than the MA5 fit at high pt, somewhat poorer at lo moderate pt


Note that the mistag rate goes up to 6% at 1000 GeV; the 1.1% quoted is because most of the jets in the sample were lower pt
    */


  for(auto jet : jets) { 
    double jpt=jet->pT();
    if (jpt > ptmin && jet->abseta()<etamax) { 
        if(jet->btag())
        {
            if(btag_rand(engine) < deepcsv_btag_medium(jpt)) return true; 
        }
        else
        {
             if(btag_rand(engine) <deepcsv_mistag_medium(jpt)) return true; 
        }
    } 
  } 
  return false; 
} 

double dphi_jet_met(vector<HEP::Jet*> jets, HEP::P4 &MET, double pt_min) { 
  double dphijm = 5; 
  for(int ijet=0; ijet < min(4, int(jets.size())); ijet++) { 
    if(jets.at(ijet)->pT() < pt_min){ 
      continue; 
    } 
    dphijm = min(dphijm, jets.at(ijet)->mom().deltaPhi(MET));  // deltaphi in mathutils ensures within 0 and PI
  } 
  return dphijm; 
} 
 
string monojet_bin_name(int ibin, int year){ 
    char name[50]; 
    if(ibin<0) { 
      sprintf(name, "monojet_SR_%d", year); 
    } else { 
      sprintf(name, "monojet_SR_%d_bin%d", year, ibin); 
    } 
    return string(name); 
} 


bool hem_veto_jets(vector<HEP::Jet*> jets) { 
  for(auto jet : jets) { 
     if(jet->pT()<30){ 
       continue; 
     } 
     if(jet->eta()<-3.0 or jet->eta() > -1.3){ 
       continue; 
     } 
     if (jet->phi()<-1.57 or jet->phi()> -0.87) { 
       continue; 
     } 
    return false; 
  } 
  return true; 
} 
/*
  In the 2018 data set, an eta-phi sector of
  the HCAL is broken. We therefore have to
  veto any event with where MET phi points 
  in this direction.
*/
bool hem_veto_met(double met_pt, double met_phi) { 
  return (met_pt > 470) or (met_phi > -0.62) or (met_phi<-1.62); 
} 


    }
}

using namespace HEP::CMS_EXO_20_004;

void CMS_EXO_20_004::init() {

cout << "---------------------------------------------------" << endl;
cout << "-- Analysis CMS_EXO_20_004 --" << endl;
cout << "---------------------------------------------------" << endl;


vector<string> allSRs;

const std::vector<std::string> SRyears= {"monojet_SR_2016","monojet_SR_2017","monojet_SR_2018"};
//cout << "Start adding regions" << endl;
/*
for(auto year : {2016, 2017, 2018}){ 
AddRegionSelection(monojet_bin_name(-1, year)); 
 
    for(size_t ibin=0; ibin<monojet_bins.size()-1; ibin++) { 
AddRegionSelection(monojet_bin_name(int(ibin), year)); 
    }

  } 
*/

for(auto year : {2016, 2017, 2018}){ 

    for(size_t ibin=0; ibin<monojet_bins.size()-1; ibin++) {
        string tname= monojet_bin_name(int(ibin), year);
        allSRs.push_back(tname);
        AddRegionSelection(tname); 
    }

  } 



vector<string> cut_order; 
  vector<string> common = { 
    "ptmiss", 
    "veto_electron", 
    "veto_muon", 
    "veto_tau", 
    "veto_btag", 
    "veto_photon", 
    "dphijm", 
    "leadak4_pt", 
    "leadak4_eta", 
  }; 


 // cout << "Start adding cuts " << endl;
  for(auto const & cut : common) { 
AddCut(cut,allSRs); 
    cut_order.push_back(cut); 
  } 
 

// HEM veto is for 2018 only 
  const std::vector<std::string> regions_2018 = { 
    //"monojet_SR_2018", 
    "monojet_SR_2018_bin0", 
    "monojet_SR_2018_bin1", 
    "monojet_SR_2018_bin2", 
    "monojet_SR_2018_bin3", 
    "monojet_SR_2018_bin4", 
    "monojet_SR_2018_bin5", 
    "monojet_SR_2018_bin6", 
    "monojet_SR_2018_bin7", 
    "monojet_SR_2018_bin8", 
    "monojet_SR_2018_bin9", 
    "monojet_SR_2018_bin10", 
    "monojet_SR_2018_bin11", 
    "monojet_SR_2018_bin12", 
    "monojet_SR_2018_bin13", 
    "monojet_SR_2018_bin14", 
    "monojet_SR_2018_bin15", 
    "monojet_SR_2018_bin16", 
    "monojet_SR_2018_bin17", 
    "monojet_SR_2018_bin18", 
    "monojet_SR_2018_bin19", 
    "monojet_SR_2018_bin20", 
    "monojet_SR_2018_bin21", 
  }; 
 //  cout << "hem ehem " << endl;
AddCut("hem_veto_jets",  regions_2018); 
  cut_order.push_back(string("hem_veto_jets")); 
 
AddCut("hem_veto_met",  regions_2018); 
  cut_order.push_back(string("hem_veto_met")); 
 


 //cout << "Start adding one cut per region " << endl;
  // Create region selections 
  for(auto year : {2016,2017,2018}) { 
    for(size_t ibin=0; ibin<monojet_bins.size()-1; ibin++) { 
      // Region specific cuts 
      auto region_name = monojet_bin_name(ibin, year); 
AddCut(region_name,  region_name); 
      cut_order.push_back(region_name); 
    } 
  } 
 
  
  // Sanity check: No duplicate cut definitions 
  //set<string> set_of_cuts( cut_order.begin(), cut_order.end() ); 
  //assert(set_of_cuts.size() == cut_order.size()); 

cout << "Initialised!" << endl;
//this->DetectorFunction="DEFAULT";
this->DetectorFunction="ATLAS";
}

CMS_EXO_20_004::~CMS_EXO_20_004() {

};

void CMS_EXO_20_004::Execute(std::mt19937 &engine) {


 // Cut definitions 
  double const CUT_ELECTRON_PT_MIN  = 10; 
  double const CUT_ELECTRON_ETA_MAX = 2.4; 
 
  double const CUT_MUON_PT_MIN      = 10; 
  double const CUT_MUON_ETA_MAX     = 2.5; 
 
  double const CUT_TAU_PT_MIN      = 18; 
  double const CUT_TAU_ETA_MAX     = 2.3; 
 
  double const CUT_PHO_PT_MIN      = 15; 
  double const CUT_PHO_ETA_MAX     = 2.5; 
   
  double const CUT_BTAG_PT_MIN      = 20; 
  double const CUT_BTAG_ETA_MAX     = 2.4; 
 
  double const LEADAK4_ETA_MAX = 2.4; 
  double const LEADAK4_PT_MIN = 100; 
  double const DPHI_JET_MET_MIN = 0.5; 
    // Filtered collections 

  
vector<HEP::Particle*> electrons = Event->electrons();
filterPhaseSpace(electrons, CUT_ELECTRON_PT_MIN, CUT_ELECTRON_ETA_MAX); 
filterEfficiency(electrons,&HEP::CMS_EXO_20_004::electronRecoEff,engine);
//filterObjects(electrons,this->Event,engine,&HEP::CMS_EXO_20_004::electronIsolation);

vector<HEP::Particle*> muons     = Event->muons();
filterPhaseSpace(muons,     CUT_MUON_PT_MIN,     CUT_MUON_ETA_MAX    ); 
filterEfficiency(muons,&HEP::CMS_EXO_20_004::muonRecoEff,engine);
//filterObjects(muons,this->Event,engine,&HEP::CMS_EXO_20_004::muonIsolation);


vector<HEP::Particle*> taus      = Event->taus();
filterPhaseSpace(taus,      CUT_TAU_PT_MIN,      CUT_TAU_ETA_MAX     ); 

vector<HEP::Particle*> photons   = Event->photons();
filterPhaseSpace(photons,   CUT_PHO_PT_MIN,      CUT_PHO_ETA_MAX     ); 
filterEfficiency(photons,&HEP::CMS_EXO_20_004::electronRecoEff,engine);  // formula from CMS is identical to electrons ...
//filterObjects(photons,this->Event,engine,&HEP::CMS_EXO_20_004::electronIsolation);
  vector<HEP::Jet*> jets      = Event->jets();
  filterPhaseSpace(jets, 20.0, 4.8); 
    /*
  vector<HEP::Particle> electrons = filter_collection<RecLeptonFormat>(Event->electrons(), CUT_ELECTRON_PT_MIN, CUT_ELECTRON_ETA_MAX); 
  vector<HEP::Particle> muons     = filter_collection<RecLeptonFormat>(Event->muons(),     CUT_MUON_PT_MIN,     CUT_MUON_ETA_MAX    ); 
  vector<HEP::Particle> taus      = filter_collection<RecTauFormat   >(Event->taus(),      CUT_TAU_PT_MIN,      CUT_TAU_ETA_MAX     ); 
  vector<HEP::Particle> photons   = filter_collection<RecPhotonFormat>(Event->photons(),   CUT_PHO_PT_MIN,      CUT_PHO_ETA_MAX     ); 
  vector<HEP::Jet> jets      = filter_collection<RecJetFormat   >(Event->jets(),      20.0,                999                 ); 
    */
   
  // Quantities to cut on 
  auto metvec=Event->missingmom();
  double met_pt = metvec.pT(); 
  double met_phi = metvec.phi(); 
  double dphijm      = dphi_jet_met(jets, metvec, 30); 
  double leadak4_pt  = jets.size() > 0 ? jets.at(0)->pT() : 0; 
  double leadak4_eta = jets.size() > 0 ? jets.at(0)->eta() : -999; 
  bool veto_btag = not has_b_tag(jets, CUT_BTAG_PT_MIN, CUT_BTAG_ETA_MAX,engine); 
  
 
  // Histogram filling 
  // Manager()->FillHisto("ptmiss", event.rec()->MET().pt() ); 
  // Cut application 
ApplyCut(met_pt>250,       "ptmiss"); 
ApplyCut(electrons.size() == 0,             "veto_electron"); 
ApplyCut(muons.size() == 0,                 "veto_muon"); 

// could apply a flat 60% efficiency on tau tagging as per CMS version of MA5, but actually the ma5 analysis 
// just demands taus. However since these come from delphes, I think it refers only to 
// hadronically decaying taus ... in any case, for this analysis leptonic decays would already be picked up from
// the above cuts
// 

/*
Hadronically decaying τ leptons are identiﬁed using the “hadrons-plus-strips” algorithm and
a multivariate classiﬁer at a working point corresponding to an efﬁciency of 70% for genuine
7
τ decays and 0.5–3% for jets from QCD production, depending on jet pT [69] https://arxiv.org/pdf/1809.02816.pdf

see table 12 page 33

p12:
The working points for the reconstructed τh
candidates are chosen to have isolation efﬁciencies between 40 and 90%, in steps of 10%, for
the reconstructed τh candidates

These are VLoose, Loose, Medium, Tight, VTight, VVTight

so we want the 'Tight' working point

The CMS analysis uses a flat 1% mistag rate for jets, this is only true for jets around 40-50 GeV in pT


*/
if((met_pt < 250) || (electrons.size()>0) ||(muons.size()>0) ) return;

int nfoundtaus=0;
for (auto tau: taus)
{
   // if(rand_dist(engine) < 0.7) nfoundtaus++;
  // flat 70% rate but on 65% of taus (which decay hadronically):
  if(rand_dist(engine) < 0.46) nfoundtaus++;

}
//for (auto jet: jets)


ApplyCut(nfoundtaus == 0,                  "veto_tau");


//ApplyCut(taus.size() == 0,                  "veto_tau"); 
ApplyCut(veto_btag,                         "veto_btag"); 
ApplyCut(photons.size() == 0,               "veto_photon"); 
ApplyCut(dphijm>DPHI_JET_MET_MIN,           "dphijm"); 
ApplyCut(leadak4_pt>LEADAK4_PT_MIN,         "leadak4_pt"); 
ApplyCut(fabs(leadak4_eta)<LEADAK4_ETA_MAX, "leadak4_eta"); 
 
  // HEM veto (2018 only) 
  ApplyCut(hem_veto_jets(jets), "hem_veto_jets"); 
ApplyCut(hem_veto_met(met_pt, met_phi), "hem_veto_met"); 



  // Verify that the random number will be uniform in [0,1] 
  //assert(rand_dist.a()==0); 
  //assert(rand_dist.b()==1); 
  double rand = rand_dist(engine); 
  for(auto year : {2016, 2017, 2018}) { 
    // Use random number for lumi weighting 
    // All regions will be normalized by MadAnalysis to 137 fb-1 
    // Here, we implement a per-region weight that corrects this 
    // down to the per-year luminosity. 
    // Each event is assigned sa either 2016, 2017 or 2018, 
    // so that there is no stat. overlap and as many generated 
    // events as possible are used. 
    bool pass_lumi_scaling = false; 
    if(year == 2016) { 
      // 36 fb-1 
      pass_lumi_scaling = rand <= 36. / 137; 
    } else if (year ==2017) { 
      // 41 fb-1 
      pass_lumi_scaling = rand > 36. / 137; 
      pass_lumi_scaling &= rand <= (36.+41.) / 137; 
    } else { 
      // 60 fb-1 
      pass_lumi_scaling = rand > (36.+41.) / 137; 
      pass_lumi_scaling &= rand <= 1; 
    } 
 
    for(size_t ibin=0; ibin<monojet_bins.size()-1; ibin++){ 
      auto cut_name = monojet_bin_name(ibin, year); 
      bool pass = (met_pt > monojet_bins.at(ibin)) && (met_pt < monojet_bins.at(ibin+1)) && pass_lumi_scaling; 
ApplyCut(pass, cut_name); 
    } 
  } 
   


}

