"""
Package to invoke the statistics check for a given analysis

Use either pyhf, native or spey. Don't touch madanalysis

"""


try:
    from lxml import ET
except Exception as err:

    try:
        import xml.etree.ElementTree as ET
    except Exception as err:
        print('lxml or xml not available... the ma5 module cannot be used')
        print(str(err))

import pkgutil

try:
    from importlib import resources as impresources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as impresources



import Data as HAData


import src.HApyhf
import src.HAspey
import src.HAnative
import json
import zslha
"""
Read a json file specifying:
- Analysis name
- Lumi
- Type (pyhf/native/spey/???)
- Filename of the stats info (Bkg json for pyhf; xml file for madanalysis/spey; json for native)
- Dictionary of cuts in the stats source to those in the analysis

"""



class AnalysisStats:
    def __init__(self,indict,choose_type=None):
        self.analysis=indict['Analysis']
        self.type = indict['Type']
        if choose_type is not None:
            self.type = choose_type
        if 'Lumi' in indict:
            self.lumi=float(indict['Lumi'])
        else:
            self.lumi=None
        #self.jsonname=indict['input_json_filename']
        if self.type == 'pyhf':
            self.jsonname=indict['Background File']
            #thispyhfjson=pkgutil.get_data(__name__, "Data/"+self.jsonname)
            #thispyhf=json.load(thispyhfjson)
            with impresources.open_text(HAData,self.jsonname) as thisinfo:
                thispyhf=json.load(thisinfo)

            self.Handler=src.HApyhf.PyhfData(thispyhf,indict['Region Dictionary'])
        elif self.type == 'MA5spey':
            self.infoname=indict['Info File']
            ma5analysis=self.analysis.lower()
            #info_input=pkgutil.get_data(__name__, "Data/"+self.infoname)
            #info_tree = ET.parse(info_input)
            with impresources.open_text(HAData,self.infoname) as info_input:
                info_tree = ET.parse(info_input)

            self.Handler=src.HAspey.HandleMA5(info_tree,ma5analysis)
        elif self.type == 'HAspey':
            print(f"{self.type} not yet implemented")
        elif self.type == 'Native':
            """
            jsonname=indict['HA File']
            with impresources.open_text(HAData,jsonname) as thisinfo:
                thisnative=json.load(thisinfo)
            """
            self.Handler=src.HAnative.HAnative(indict)
            #print(f"{self.type} not yet implemented")
        else:
            raise NameError("Did not recognise Analysis type")





    def getstatsfromdict(self,effdict,inputxs=None,with_significance=True,with_upper_limits=False):
        """
        Get stats after providing dict in HackAnalysis format for the specific analysis
        Should contain a set of efficiencies
        """
        results={}
        analysis=self.analysis.upper()
        if inputxs is None:
            print('No xs provided')
            if 'xs' not in effdict:
                raise NameError("No cross-section in dictionary and none provided")
            xs=float(effdict['xs']) ### HA gives in fb
        else:
            xs=float(inputxs)
        efficiencies={}
        signal_data={}
        
        #lumi=self.lumi
        for region in effdict["Regions"]:
            
            efficiencies[region]=effdict["Regions"][region]["efficiency"]
            if self.lumi is not None:
                signal_data[region]=efficiencies[region]*self.lumi*xs
        if self.type == 'pyhf':
            print("Calling the pyhf handler")
            results=self.Handler.getCLs(signal_data,with_significance,with_upper_limits)
        elif self.type == 'MA5spey' or self.type=='HAspey':
            print("Calling the spey handler")
            results=self.Handler.runspeycombined(efficiencies,xs,with_significance)
        elif self.type == 'Native':
            print("Calling the native handler")
            results=self.Handler.getlimits(efficiencies,xs,with_upper_limits)
        else:
            results={}
        return results
        

    def getstats(self,spcname,inputxs=None,with_significance=True,with_upper_limits=False):
        HAeff=zslha.read(spcname)
        results={}
        analysis=self.analysis.upper()
        
        if inputxs is None:
            print('No xs provided')
            xs=float(HAeff.analyses[analysis]['PROCESS']['XS']) ### HA gives in fb
        else:
            xs=float(inputxs)
    
        efficiencies={}
        signal_data={}
        
        #lumi=self.lumi
        for region in HAeff.analyses[analysis]['EFFICIENCIES']:
            
            eff=HAeff.Value('EFFICIENCIES',[analysis,region,0])
            efficiencies[region]=eff
            if self.lumi is not None:
                signal_data[region]=efficiencies[region]*self.lumi*xs
        if self.type == 'pyhf':
            print("Calling the pyhf handler")
            results=self.Handler.getCLs(signal_data,with_significance,with_upper_limits)
        elif self.type == 'MA5spey' or self.type=='HAspey':
            print("Calling the spey handler")
            results=self.Handler.runspeycombined(efficiencies,xs,with_significance)
        elif self.type == 'Native':
            print("Calling the native handler")
            results=self.Handler.getlimits(efficiencies,xs,with_upper_limits)
        else:
            results={}
        return results
            

if __name__ == '__main__':
    import sys,os
    if len(sys.argv) < 2:
        print("Usage: \n HAStats.py <HA efficiency file> [Optional: cross-section] [Optional: path for Analysis info files]")
        raise SystemExit
    
    spcname=sys.argv[1]
    if not os.path.exists(spcname):
        print('Could not open efficiency file!')
        raise SystemExit
    MAINpath = os.path.dirname(os.path.realpath( __file__ ))
    sys.path.insert(0,MAINpath)

    if len(sys.argv) > 2:
        inputxs=float(sys.argv[2])
    else:
        inputxs=None
    if len(sys.argv) > 3:
        infopath=sys.argv[3]
        datapath=sys.argv[3]
    else:
        infopath=os.path.join(MAINpath,"Analyses")
        datapath=os.path.join(MAINpath,"Data")

    import zslha

    import Analyses as HAAnalyses
    HAeff=zslha.read(spcname)

    results={}

    for analysis in HAeff.analyses:
        infojsonname=analysis+'_info.json'
        headstring="====================== %s ================" %analysis
        print(headstring)
        try:
            with impresources.open_text(HAAnalyses,infojsonname) as thisinfo:
                analysis_infodict=json.load(thisinfo)
        except Exception as e:
            print(f"Failed to initialise analysis {analysis}:"+str(e))
            continue
        handler=AnalysisStats(analysis_infodict)
        results = handler.getstats(spcname,inputxs=inputxs,with_significance=True)
           
        for key, val in results.items():
            print(key+": "+str(val))
        print("----------------------------------------------------")
        """
        if handler.type=='pyhf':

        elif handler.type =='MA5spey':
        """

    """
    for analysis in HAeff.analyses:
        if inputxs is not None:
            xs=float(HAeff.analyses[analysis]['PROCESS']['XS']) ### HA gives in fb
        else:
            xs=float(inputxs)
        efficiencies={}
        
        for region in HAeff.analyses[analysis]['EFFICIENCIES']:
            eff=HAeff.Value('EFFICIENCIES',[analysis,region,0])
            efficiencies[region]=eff
            

        infojsonname=analysis+'_info.json'
        ma5analysis=analysis.lower()
        ma5infoname=ma5analysis+'.info'
        if os.path.exists(os.path.join(infopath,infojsonname)):
            headstring="====================== %s ================" %analysis
            print(headstring)
            print("Reading json file!")
            with open(os.path.join(infopath,infojsonname),'r') as json_info:
                info_dict=json.load(json_info)
            lumi=float(info_dict['Lumi'])
            
            with open(os.path.join(datapath,info_dict['Background File']),'r') as pyhf_json:
                thispyhf=json.load(pyhf_json)
            signal_data={}
            for region in efficiencies:
                signal_data[region]=efficiencies[region]*lumi*xs

            thisHandler=src.HApyhf.PyhfData(thispyhf,info_dict['Region Dictionary'])

            results=thisHandler.getCLs(signal_data)
            
            print("CLs observed: %.4f" %results['CLs obs'])
            print("CLs expected: "+str(results["CLs exp"]))
            print("pval observed: %.4e" %results['pval'])
            print("pval expected: " +str(results['pval exp']))
            print("="*len(headstring))

        elif os.path.exists(os.path.join(datapath,ma5infoname)):
            headstring="====================== %s ================" %ma5analysis
            print(headstring)
            print("Reading MA5 info file!")

            with open(os.path.join(datapath,ma5infoname), "r") as info_input:
                info_tree = ET.parse(info_input)
            thisHandler=src.HAspey.HandleMA5(info_tree,ma5analysis)
            results=thisHandler.runspeycombined(efficiencies,xs)
            ## 'mumax':mumax, 'sigmamu':sigmamu, 'pval':pval, 'uppermuexp':uppermuexp, 'uppermuobs':uppermuobs
            
            print("xs: %.4e"%xs)
            print("upper mu expected: %.4f"%results['uppermuexp'])
            print("upper mu observed: %.4f"%results['uppermuobs'])
            print("pval: %.4f"%results['pval'])
            print("="*len(headstring))
        else:
            print('Could not find info file for the required analysis! Must have the name <analysisname>_info.json or <analysisname>.info for MA5 covariance-style analyses')
            continue

"""