import jsonpatch
import jsonschema
import json
import os

import numpy as np


import time, sys,math


try:
    import pyhf
except ModuleNotFoundError:
    print ( "pyhf import failed. Is the module installed?" )
    sys.exit(-1)

try:
    pyhf.set_backend("pytorch")
    
except Exception as e:
    print("Could not set pytorch as backend: "+str(e))
    pyhf.set_backend("numpy")
 
print(pyhf.get_backend())





   
class PyhfData:
    """
    Holds data for use in pyhf
    :ivar nsignals: signal predictions list divided into sublists, one for each json file
    :ivar inputJsons: list of json instances
    :ivar nWS: number of workspaces = number of json files
    """
    def __init__ (self, inputJson, pyhf_to_cut_dict={}):
        self.signals = []
        self.inputJson = inputJson
        self.pyhf_cut_dict = pyhf_to_cut_dict
        self.getWSInfo()
        self.msettings = {'normsys': {'interpcode': 'code4'},'histosys': {'interpcode': 'code4p'}}

        #for srInfo in self.channelInfo['signalRegions']:
        #    print('SR: %s' %srInfo)
        
    def getWSInfo(self):
        """
        Get information from the json file

        :ivar channelsInfo: list of dictionaries (one dictionary for each json file) containing useful information about the json files
            - :key signalRegions: list of dictonaries with 'json path' and 'size' (number of bins) of the 'signal regions' channels in the json files
            - :key otherRegions: list of strings indicating the path to the control and validation region channels
        """
        # Identifying the path to the SR and VR channels in the main workspace files
        self.channelsInfo = [] # workspace specifications
        
        ws=self.inputJson
        wsChannelsInfo = {}
        wsChannelsInfo['signalRegions'] = []
        wsChannelsInfo['otherRegions'] = []
        if not 'channels' in ws.keys():
            self.channelInfo = None
            return
        self.channelInfo=[]
        #index_to_cut={}
        for i_ch, ch in enumerate(ws['channels']):
            #print(ch['name'])
            if ch['name'][:2] == 'SR': # if channel name starts with 'SR'
                #if ch['name'] not in self.pyhf_cut_dict:
                #    self.pyhf_cut_dict[ch['name']]=ch['name']
                
                #cuts_to_path[ch['name']] = '/channels/'+str(i_ch)+'/samples/0'
                nsamples=len(ch['samples'])  ### original from smodels used 0 here instead
                #print(ch['name']+': '+str(len(ch['samples'][0]['data']))+': '+'/channels/'+str(i_ch)+'/samples/'+str(nsamples))
                
                #print(ch)
                wsChannelsInfo['signalRegions'].append({'cutname': ch['name'],'path':'/channels/'+str(i_ch)+'/samples/'+str(nsamples), # Path of the new sample to add (signal prediction)
                                                            'size':len(ch['samples'][0]['data'])}) # Number of bins
                #wsChannelsInfo['signalRegions'].append({'path':'/channels/'+str(i_ch)+'/samples/'+str(nsamp), # Path of the new sample to add (signal prediction)
                #                                            'size':len(ch['samples'][0]['data'])}) # Number of bins
                #nsamp=nsamp+1
            else:
                #print('Other: '+ch['name'])
                wsChannelsInfo['otherRegions'].append('/channels/'+str(i_ch))
        wsChannelsInfo['otherRegions'].sort(key=lambda path: path.split('/')[-1], reverse=True) # Need to sort correctly the paths to the channels to be removed
        self.channelInfo = wsChannelsInfo
        #print('SR Channels: '+str(wsChannelsInfo['signalRegions']))
        #print('Other regions: '+str(wsChannelsInfo['otherRegions']))



    def patchMaker(self,signaldata):
        """
        Method that creates the list of patches to be applied to the `self.inputJsons` workspaces, one for each region given the `self.nsignals` and the informations available in `self.channelsInfo` and the content of the `self.inputJsons`
        NB: It seems we need to include the change of the "modifiers" in the patches as well

        :return: the list of patches, one for each workspace
        """
        if self.channelsInfo == None:
            return None
        
        # Constructing the patches to be applied on the main workspace files

        
        patch = []
        #subSig=signals
        for srInfo in self.channelInfo['signalRegions']:
            """
            In HackAnalysis each region maps to one bin, but in a pyhf file there may be multiple samples per SR
            """
            nBins = srInfo['size']
            operator = {}
            operator["op"] = "add"
            operator["path"] = srInfo['path']
            value = {}
            #value["data"] = subSig[:nBins]  ## take first nBins from the list of signals
            #subSig = subSig[nBins:]
            PYHF_name=srInfo["cutname"]
            if self.pyhf_cut_dict=={}:
                HA_name=PYHF_name
                value["data"] = [signaldata[HA_name]]
            else:
                if int(nBins) > 1:
                    data=[]
                    for nbin in range(nBins):
                        pycutname=PYHF_name+'_'+str(nbin)
                        HA_name=self.pyhf_cut_dict[pycutname]
                        data.append(signaldata[HA_name])
                    value["data"] = data
                else:
                    HA_name=self.pyhf_cut_dict[PYHF_name]
                    value["data"] = [signaldata[HA_name]]
            #print('%s NAME: %s, value: %s'%(PYHF_name,HA_name,str(value)))
            
            value["modifiers"] = []
            value["modifiers"].append({"data": None, "type": "normfactor", "name": "mu_SIG"})
            value["modifiers"].append({"data": None, "type": "lumi", "name": "lumi"})
            value["name"] = "bsm"
            operator["value"] = value
            patch.append(operator)

        ## really? NO!!!
        
        #for path in self.channelInfo['otherRegions']:
        #    patch.append({'op':'remove', 'path':path})
        
        return patch

    def wsMaker(self,signals):
        """
        Apply each region patch (self.patches) to his associated json (self.inputJsons) to obtain the complete workspaces

        :returns: the list of patched workspaces
        """
        patch=self.patchMaker(signals)
        #print(patch)
        try:
            newjs=jsonpatch.apply_patch(self.inputJson, patch)
        except:
            print('failed to create patch')
            return None
        try:
            newws=pyhf.Workspace(newjs)
            return newws
        except (pyhf.exceptions.InvalidSpecification, KeyError) as e:
            print('Error in workspace patch making: '+str(e))
            return None


    def getCLs(self,signaldata,with_significance=True,getmumax=False):
        #print(signaldata)
        results={}
        try:
            wspace = self.wsMaker(signaldata)
            model = wspace.model(measurement_name=None, modifier_settings=self.msettings)
            data = wspace.data(model)
            
            """
            nb in pyhf infer/__init__.py"
            
            is_q0 = kwargs.get('test_stat', 'qtilde') == 'q0'

                _returns = [CLsb_obs if is_q0 else CLs_obs]
            if return_tail_probs:
                if is_q0:
                    _returns.append([CLb_obs])
                else:
                    _returns.append([CLsb_obs, CLb_obs])
            
            """
            if not with_significance:
                #CLs_obs, CLs_exp = pyhf.infer.hypotest(1.0, data, model, return_expected_set=True, test_stat="qtilde",calctype="toybased",ntoys=1000,track_progress=True)
                CLs_obs, CLs_exp = pyhf.infer.hypotest(1.0, data, model, return_expected_set=True, test_stat="qtilde")
                results={'CLs obs':float(CLs_obs), 'CLs exp': list(map(float,CLs_exp))}

            else:
                #parnames=model.config.par_names

                ## index of poi:
                poiind=model.config.poi_index
                #poiname=model.config.poi_name
                asymptotic_calculator = pyhf.infer.utils.create_calculator("asymptotics", data, model, test_stat="qtilde")
                mu_test=1.0
                q_tilde = asymptotic_calculator.teststatistic(mu_test)

                sig_plus_bkg_dist, bkg_dist = asymptotic_calculator.distributions(mu_test)

                CLsb, CLb, CLs = asymptotic_calculator.pvalues(q_tilde, sig_plus_bkg_dist, bkg_dist)
                CLsb_exp_band, CLb_exp_band, CLs_exp_band = asymptotic_calculator.expected_pvalues(sig_plus_bkg_dist, bkg_dist)
                #fitpars=asymptotic_calculator.fitted_pars
                bestmu=asymptotic_calculator.fitted_pars.free_fit_to_data[poiind]

                resCLs_obs=float(CLs)
                resCLs_exp=list(map(float,CLs_exp_band))[2]
                results={'CLs obs':resCLs_obs, 'CLs exp': list(map(float,CLs_exp_band))}
                results['bestmu']=float(bestmu)
                if getmumax:
                    ## do for expected and observed
                    def getclsobs(mu):
                        q_tilde = asymptotic_calculator.teststatistic(mu)
                        sig_plus_bkg_dist, bkg_dist = asymptotic_calculator.distributions(mu)
                        CLsb, CLb, CLs = asymptotic_calculator.pvalues(q_tilde, sig_plus_bkg_dist, bkg_dist)
                        #print(f"mu: {mu}, CLs: {CLs}")
                        return float(CLs)-0.05
                
                    def getclsexp(mu):
                        q_tilde = asymptotic_calculator.teststatistic(mu)
                        sig_plus_bkg_dist, bkg_dist = asymptotic_calculator.distributions(mu)
                        CLsb_exp_band, CLb_exp_band, CLs_exp_band = asymptotic_calculator.expected_pvalues(sig_plus_bkg_dist, bkg_dist)
                        clsexpband=list(map(float, CLs_exp_band))
                        #print(f"mu: {mu}, CLs exp: {clsexpband[2]}")
                        return clsexpband[2]-0.05
                    
                    doobserved=True
                    doexpected=True
                    """
                    if resCLs_obs < 0.005: # give up
                        mu95obs=0.0
                        doobserved=False
                    elif resCLs_obs > 0.995:
                        mu95obs=10.0
                       
                        doobserved=False
                    elif resCLs_obs < 0.05:
                        obsmin=0.1
                        obsmax=1
                    elif resCLs_obs < 0.5:
                        obsmin=1
                        obsmax=5
                    else:
                        obsmin=1
                        obsmax=10
                    from scipy.optimize import brentq
                    if doobserved:
                        try:
                            #mulimobs=brentq(getclsobs,obsmin,obsmax,maxiter=10)
                            mu95obs=brentq(getclsobs,obsmin,obsmax,rtol=0.01) ## allow percent-level uncertainty
                        except:
                            if obsmin < 1:
                                mu95obs = 0.0
                            else:
                                mu95obs = 10.0
                    if resCLs_exp < 0.005: # give up
                        mu95exp=0.0
                        doexpected=False
                    elif resCLs_exp > 0.995:
                        mu95exp=10.0
                       
                        doexpected=False
                    elif resCLs_exp < 0.05:
                        expmin=0.1
                        expmax=1
                    elif resCLs_exp < 0.5:
                        expmin=1
                        expmax=5
                    else:
                        expmin=1
                        expmax=10
                    
                    if doexpected:
                        try:
                            #mulimobs=brentq(getclsobs,obsmin,obsmax,maxiter=10)
                            mu95exp=brentq(getclsexp,expmin,expmax,rtol=0.01) ## allow percent-level uncertainty
                        except:
                            if expmin < 1:
                                mu95exp = 0.0
                            else:
                                mu95exp = 10.0
                    """
                    if resCLs_obs < 0.05:
                        obsmax=1.0
                        obsmin=1.0
                        tempcls=resCLs_obs-0.05
                        lastmin=obsmax
                        try:
                            while tempcls <= 0.0:
                                obsmax=lastmin
                                lastmin=obsmin
                                obsmin=obsmin*0.2
                                tempcls=getclsobs(obsmin)
                        except:
                            mu95obs=0.0
                            doobserved=False
                    else:
                        obsmin=1.0
                        obsmax=1.0
                        tempcls=resCLs_obs-0.05
                        lastmax=obsmin
                        try:
                            while tempcls >= 0.0:
                                obsmin=lastmax
                                lastmax=obsmax
                                obsmax=obsmax*5.0
                                tempcls=getclsobs(obsmax)
                        except:
                            mu95obs=10.0
                            doobserved=False
                    if resCLs_exp < 0.05:
                        expmax=1.0
                        expmin=1.0
                        tempcls=resCLs_exp-0.05
                        lastmin=expmax
                        try:
                            while tempcls <= 0.0:
                                expmax=lastmin
                                lastmin=expmin
                                expmin=expmin*0.2
                                tempcls=getclsexp(expmin)
                        except:
                            mu95exp=0.0
                            doexpected=False
                    else:
                        expmin=1.0
                        expmax=1.0
                        tempcls=resCLs_exp-0.05
                        lastmax=expmin
                        try:
                            while tempcls >= 0.0:
                                expmin=lastmax
                                lastmax=expmax
                                expmax=expmax*5.0
                                tempcls=getclsexp(expmax)
                        except:
                            mu95exp=10.0
                            doexpected=False
                    from scipy.optimize import brentq
                    if doobserved:
                        try:
                            mu95obs=brentq(getclsobs,obsmin,obsmax,rtol=0.01) ## allow percent-level uncertainty
                        except:
                            if obsmin < 1:
                                mu95obs = 0.0
                            else:
                                mu95obs = 10.0

                    if doexpected:
                        try:
                            mu95exp=brentq(getclsexp,expmin,expmax,rtol=0.01) ## allow percent-level uncertainty
                        except:
                            if expmin < 1:
                                mu95exp = 0.0
                            else:
                                mu95exp = 10.0

                    results['mu95obs']=mu95obs
                    results['mu95exp']=mu95exp
                pval_obs, CLb, pval_exp = pyhf.infer.hypotest(0.0, data, model, return_tail_probs=True, return_expected_set=True, test_stat="q0")
                results['pval']=pval_obs
                results['pval exp']=pval_exp
        except Exception as e:
            print('Failure: '+str(e))
            return {}
        


        #print("CLb: "+str(CLb))
        #results={'CLs obs':float(CLs_obs), 'CLs exp': list(map(float,CLs_exp)), 'pval':pval_obs, 'pval exp':pval_exp}
        return results