#pragma once

/*

General routines for building your pythia event fill function



*/


HEP::P4 PVtoH(const Pythia8::Vec4 &pyv)
{
  HEP::P4 newv(pyv.px(), pyv.py(), pyv.pz(), pyv.e());
  return newv;
  //return HEP::P4(double pyv.px(), double pyv.py(), double pyv.pz(), double pyv.E());
}


Pythia8::Vec4 getMissingMomentum(Pythia8::Event &event)
{
  Pythia8::Vec4 missingETvec;
  std::vector<int> metaStableMothers;
    //First check for all final particles:
    for (int i = 0; i < event.size(); ++i) {
    	Pythia8::Particle ptc = event[i];
    	if (!ptc.isFinal()) continue;
    	//If particle has been produced inside or after the calorimeter, skip it and include its parent
      // These for ATLAS
    	//if (abs(ptc.vProd().pz()) > 6100. || ptc.vProd().pT() > 2300.){
        // These for CMS
        //if (abs(ptc.vProd().pz()) > 5500. || ptc.vProd().pT() > 2950.){
          // CMS muon chamber. Assuming here something invisible produced by something charged
          if (fabs(ptc.vProd().pz()) > 11000. || ptc.vProd().pT() > 7000.){
    		int mom = ptc.mother1();
    		//cout << "Adding mom " << mom << " from " << ptc.name() << " and index " << i << endl;
    		if(std::find(metaStableMothers.begin(), metaStableMothers.end(), mom) == metaStableMothers.end()) {
    			metaStableMothers.push_back(mom);
    		}
    		continue;
      	}
      
      
    	//Neutrinos are included in MET:
    	if (ptc.idAbs() == 12 || ptc.idAbs() == 14 || ptc.idAbs() == 16){
    		//cout << "Including: " << i << " " << ptc.name() << endl;
    		missingETvec += ptc.p();
    		continue;
    	}
    	if (ptc.isHadron()) continue;  //Ignore usual hadrons
    	if (ptc.idAbs() == 22) continue; //Ignore photons
      if (ptc.chargeType() !=0) continue; // Ignore anything charged!!
    	if (ptc.isVisible() && ptc.m() < 10.) continue; //Ignore light visible particles

        //Add to missing momentum
    	//cout << "Including: " << i << " " << ptc.name() << endl;
        missingETvec += ptc.p();
    }

    //Now check all meta-stable mothers:
    for (int imom = 0; imom < metaStableMothers.size(); ++imom) {
    	Pythia8::Particle ptc = event[metaStableMothers[imom]];
      
      
    	//Neutrinos are included in MET:
    	if (ptc.idAbs() == 12 || ptc.idAbs() == 14 || ptc.idAbs() == 16){
    		//cout << "Including mom: " << metaStableMothers[imom] << " " << ptc.id() << endl;
    		missingETvec += ptc.p();
    		continue;
    	}
    	if (ptc.isHadron()) continue;  //Ignore usual hadrons
    	if (ptc.idAbs() == 22) continue; //Ignore photons
       if (ptc.chargeType() !=0) continue; // Ignore anything charged!!
    	if (ptc.isVisible() && ptc.m() < 10.) continue; //Ignore light visible particles

        //Add to missing momentum
    	//cout << "Including mom: " << metaStableMothers[imom] << " " << ptc.id() << endl;
        missingETvec += ptc.p();
    }

    return missingETvec;


}

void simplecalcMissingMomentum(HEP::Event &OutEvt, std::mt19937 &engine)
{
  // look at the invisibles. Bit stupid I know but whatever. 
  HEP::P4 pmiss;
  pmiss.clear();
  for(HEP::Particle* invisible : OutEvt.invisible_particles())
  {
    pmiss+=invisible->mom();
  }

  OutEvt.set_missingmom(pmiss);

  // smear here ...
}



bool isLLP(Pythia8::Particle *particle, const Pythia8::Event &event)
{
 // if (particle->colType() != 0) return false;
	if (particle->isHadron()) return false;


	//Skip light particles
	if (fabs(particle->m()) < 20.) return false;

  if (particle->isFinal()) return true;

    Pythia8::Vec4 decayVertex = particle->vDec();
    if(decayVertex.pAbs() < 0.001) return false;
	//Check if it is not an intermediate state (has itself as daughter)
	std::vector<int> daughters = particle->daughterListRecursive();
	for (int i = 0; i < daughters.size(); ++i){
		int idaughter = daughters[i];
		Pythia8::Particle daughterParticle = event[idaughter];
		if (daughterParticle.id() == particle->id()) return false;
	}

	return true;

}

bool isHSCP(Pythia8::Particle *particle, const Pythia8::Event &event)
//Return True (False) if the particle is (is not) a HSCP
{

	//Skip neutral particles
	if (!particle->isCharged()) return false;
	//Skip color charged particles or hadrons
	if (particle->colType() != 0) return false;
	if (particle->isHadron()) return false;
	//Skip light particles
	if (fabs(particle->m()) < 20.) return false;
	//Check if it is not an intermediate state (has itself as daughter)
	std::vector<int> daughters = particle->daughterListRecursive();
	for (int i = 0; i < daughters.size(); ++i){
		int idaughter = daughters[i];
		Pythia8::Particle daughterParticle = event[idaughter];
		if (daughterParticle.id() == particle->id()) return false;
	}

	return true;
}


bool decayBeforeEndHcal(Pythia8::Particle &particle)
//Return True (False) if the particle decays before (after) the hadronic calorimeter
{
	if (particle.isFinal()){return false;}
    Pythia8::Vec4 decayVertex = particle.vDec();
	if (decayVertex.pT() < 3.9e3 && abs(decayVertex.pz()) < 6.1e3){return true;}

	return false;
}

bool decayInsideAtlas(Pythia8::Particle &particle)
//Return True (False) if the particle decays inside the detector
{
	if (particle.isFinal()){return false;}
    Pythia8::Vec4 decayVertex = particle.vDec();
    if (decayVertex.pT() < 12e3 && abs(decayVertex.pz()) < 23e3){return true;}

	return false;
}

bool ischargedTrackHadron(Pythia8::Particle* particle, const Pythia8::Event &event)
{
  if(!particle->isHadron()) {return false;}
  if(!particle->isCharged()) { return false;}

  Pythia8::Vec4 prodVertex = particle->vProd();
  //if( (prodVertex.pT() > 1100.) || (abs(prodVertex.pz()) > 3100.)) {return false;}
  if( (prodVertex.pT() > 200.) || (abs(prodVertex.pz()) > 270.)) {return false;}
  
  if(particle->isFinal())
  {
    return true;
  }
  else
  {
        std::vector<int> daughters = particle->daughterListRecursive();
	      for (int i = 0; i < daughters.size(); i++){
		    int idaughter = daughters[i];
		      const Pythia8::Particle* daughterParticle = &event[idaughter];
		    if (daughterParticle->id() == particle->id()) return false;
	      }
        
        if( (particle->vDec().pT() < 300.0) ){return false;}
  }

  return true;
}


/*

inline bool fromHadron(int n, const Pythia8::Event &evt) {
      // Root particle is invalid
      if (n == 0) return false;
      const Pythia8::Particle p = evt[n];
      if (p.isHadron()) return true;
      if (p.isParton()) return false; // stop the walking at the end of the hadron level
      for (int m : p.motherList()) {
        if (fromHadron(m, evt)) return true;
      }
      return false;
    }
*/

inline bool fromHadron(int n, const Pythia8::Event &evt) {
      // Root particle is invalid
      if (n == 0) return false;
      const Pythia8::Particle* p = &evt[n];
      if (p->isParton()) return false; // stop the walking at the end of the hadron level
      if (p->isHadron()) return true;
      
      
      for (int mm : p->motherList()) {
        //cout << n << ": " << p->id() << " Mother!: " << mm << endl;
        if (fromHadron(mm, evt)) return true;
      }
      return false;
    }


/*

bool registersinHCAL(Pythia8::Particle* prt)
{
  Pythia8::Vec4 pos = prt->vProd();
  

}


*/


// This function is not used.
HEP::PileupEvent* sortPileupEvent(Pythia8::Event &event)
{
  HEP::PileupEvent* newPuE = new HEP::PileupEvent();


  for(int npart = 0; npart< event.size(); npart++)
  {
    Pythia8::Particle* prt = &event[npart];
    int pid=prt->id();
    int apid=abs(pid);
    // check if is final and produced inside tracker or if it decays outside enough (not yet)
    //cout << pid << endl;
    if(prt->isFinal())
    {
      HEP::Particle* newpart = new HEP::Particle(prt->px(),prt->py(),prt->pz(),prt->e(),pid);
      // All particles other than invisibles and muons are jet constituents. Bollocks, really we only want photons
      //if( pid == 12 || pid == 14 || pid == 16 || pid == 18 || pid == 13 || pid == 11)
      

      if(prt->isHadron() && prt->isVisible())
         {
           if(prt->isCharged())
           {
             //OutEvt.add_charged_hadron(newpart->mom());
             Pythia8::Vec4 pos = prt->vProd();
             if (pos.pAbs() > 0.0001)
                  {
                    newpart->set_prod(HEP::P4(pos.px(),pos.py(),pos.pz(),pos.e()));
                  }
             newpart->set_3Q(prt->chargeType());
             
             //newPuE->add_hadron_vertex(newpart);
             if(newpart->pT() > 1.0)
             {
              newPuE->add_charged_hadron(newpart);
             }
             else
             {
               delete newpart;
             }
           }
           else
           {
             newPuE->add_neutral_hadron(newpart->mom());
             delete newpart;
           }
         }
         else // not hadron
          {
            
              // look at the vertices
              	//if ( (*p)->production_vertex() ) { // i.e. does it have one: we want the position
                  //HepMC::FourVector pos =  (*p)->production_vertex()->position();
                  Pythia8::Vec4 pos = prt->vProd();


                  if (pos.pAbs() > 0.0001)
                  {
                    newpart->set_prod(HEP::P4(pos.px(),pos.py(),pos.pz(),pos.e()));
                  }
                  newpart->set_prompt(); 
                  newpart->set_3Q(prt->chargeType());
                  newPuE->add_particle(newpart);
               // } 
            
          }   
      // don't bother with jet constituents for this analysis
      /*
      if( pid == 12 || pid == 14 || pid == 16 || pid == 18  ) // include electrons and muons, not neutrinos
      {
        continue;
      }
       if((pid > 1000000) && (pid < 9000000)) 
      {
        continue;
      }
      //HepMC::FourVector v4((*p)->momentum());
      Pythia8::Vec4 v4=prt.p();
      fastjet::PseudoJet particleTemp(v4.px(),v4.py(),v4.pz(),v4.e());
      fjInputs.push_back(particleTemp);
      */
    }
    else
    {
      // cout << "is not final ";
       if(ischargedTrackHadron(prt,event))
       {
        
        Pythia8::Vec4 pos = prt->vProd();
        Pythia8::Vec4 vdec = prt->vDec();
        if(vdec.pT() < 33.0) continue; // check to make sure it doesn't decay too soon
        HEP::Particle* newpart = new HEP::Particle(prt->px(),prt->py(),prt->pz(),prt->e(),pid);
              if (pos.pAbs() > 0.0001)
                  {
                    newpart->set_prod(HEP::P4(pos.px(),pos.py(),pos.pz(),pos.e()));
                  }
              newpart->set_decay(HEP::P4(vdec.px(),vdec.py(),vdec.pz(),vdec.e()));
             newpart->set_3Q(prt->chargeType());
            // cout << " adding charged hadron ";
             //newPuE->add_hadron_vertex(newpart);
             newPuE->add_charged_hadron(newpart);


       }
    }


  }

  return newPuE;

}

/*

NOONONONONO
  
DEFINE HERE THE MAP FROM THE NAME OF FILL FUNCTIONS TO THE ACTUAL FUNCTION:

std::map<std::string, std::function<int(int,int)>>  funcMap =
         {{ "add", add},
          { "sub", sub}
         };

HEP::Event &OutEvt, Pythia8::Event &evt, double weight, int npileups, std::vector<HEP::PileupEvent*> &pileups, std::mt19937 &engine

*/
