// -*- C++ -*-
//
// This file is part of HEPUtils -- https://bitbucket.org/andybuckley/heputils
// Copyright (C) 2013-2018 Andy Buckley <andy.buckley@cern.ch>
//
// Embedding of HEPUtils code in other projects is permitted provided this
// notice is retained and the HEPUtils namespace and include path are changed.
//

// Duly: HEPUtils namespace changed to HEP
// Some edits by Mark Goodsell: changed everything round so that we use
// (px, py, pz, E) instead of the mass, because I want to use it for spacetime vectors!
// Necessary for production/decay vertices etc
// 

#pragma once

#include "MathUtils.h"
#include "Utils.h"
#include <sstream>
#include <iostream>
#include <stdexcept>
#include <cmath>
#include "ExHEPUtils/Vectors.h"

/// @file Physics vectors stuff
/// @author Andy Buckley <andy.buckley@cern.ch>

/// Modified by Mark Goodsell <goodsell@lpthe.jussieu.fr>

namespace HEP {


class P3 {
  private:

    /// @name Storage
    //@{
    double _x, _y, _z;
    //@}


  public:

    /// @name Constructors etc.
    //@{

    /// Default constructor of a null vector
    P3()
      : _x(0), _y(0), _z(0) {  }

    /// Constructor from Cartesian/Minkowski coordinates
    P3(double px, double py, double pz) {
      setP(px, py, pz);
    }

    /// Copy constructor
    P3(const P3& v)
      : _x(v._x), _y(v._y), _z(v._z) {  }

    /// Copy assignment operator
    P3& operator = (const P3& v) {
      _x = v._x;
      _y = v._y;
      _z = v._z;
      return *this;
    }

    /// Set the components to zero
    void clear() {
      _x = 0;
      _y = 0;
      _z = 0;
    }

    //@}


    /// @name Static methods for vector making
    //@{

    /// Make a vector from (px,py,pz,E) coordinates
    static P3 mkXYZ(double px, double py, double pz) {
      return P3().setP(px, py, pz);
    }

    static P3 mkXY(double px, double py) {
      return P3().setXY(px, py);
    }

    /// Make a vector from (eta,phi,p) coordinates
    static P3 mkEtaPhiP(double eta, double phi, double P) {
      return P3().setEtaPhiP(eta, phi, P);
    }

    /// Make a vector from (eta,phi,pT) coordinates
    static P3 mkEtaPhiPt(double eta, double phi, double pt) {
      return P3().setEtaPhiPt(eta, phi, pt);
    }

    /// Make a vector from (theta,phi,energy) coordinates and the mass
    static P3 mkThetaPhiP(double theta, double phi, double P) {
      return P3().setThetaPhiP(theta, phi, P);
    }

    /// Make a vector from (theta,phi,pT) coordinates and the mass
    static P3 mkThetaPhiPt(double theta, double phi, double pt) {
      return P3().setThetaPhiPt(theta, phi, pt);
    }


    //@}


    /// @name Coordinate setters
    //@{

    /// Set the px coordinate
    P3& setPx(double px) {
      _x = px;
      return *this;
    }

    /// Set the py coordinate
    P3& setPy(double py) {
      _y = py;
      return *this;
    }

    /// Set the pz coordinate
    P3& setPz(double pz) {
      _z = pz;
      return *this;
    }


    /// Set the p coordinates and mass simultaneously
    P3& setP(double px, double py, double pz) {
      
      setPx(px); setPy(py); setPz(pz);
      return *this;
    }
    /// Alias for setP
    P3& SetXYZ(double px, double py, double pz) {
      return setP(px, py, pz);
    }

    /// set only transverse components
    P3& setXY(double px, double py) {
      setPx(px); setPy(py);
      return *this;
    }
    

    /// Set the vector state from (eta,phi,energy) coordinates and the mass
    ///
    /// eta = -ln(tan(theta/2))
    /// -> theta = 2 atan(exp(-eta))
    P3& setEtaPhiP(double eta, double phi, double P) {
      
      if (P < 0)
        throw std::invalid_argument("Negative momentum given as argument");
      const double theta = 2 * atan(exp(-eta));
      if (theta < 0 || theta > M_PI)
        throw std::domain_error("Polar angle outside 0..pi in calculation");
      setThetaPhiP(theta, phi,P);
      return *this;
    }

    /// Set the vector state from (eta,phi,pT) coordinates and the mass
    ///
    /// eta = -ln(tan(theta/2))
    /// -> theta = 2 atan(exp(-eta))
    P3& setEtaPhiPt(double eta, double phi, double pt) {
     
      if (pt < 0)
        throw std::invalid_argument("Negative transverse momentum given as argument");
      const double theta = 2 * atan(exp(-eta));
      if (theta < 0 || theta > M_PI)
        throw std::domain_error("Polar angle outside 0..pi in calculation");
      const double p = pt / sin(theta);
      
      setThetaPhiP(theta, phi, p);
      return *this;
    }

    /// Set the vector state from (theta,phi,energy) coordinates and the mass
    ///
    /// p = sqrt(E^2 - mass^2)
    /// pz = p cos(theta)
    /// pt = p sin(theta)
    P3& setThetaPhiP(double theta, double phi, double p) {
      if (theta < 0 || theta > M_PI)
        throw std::invalid_argument("Polar angle outside 0..pi given as argument");
      
      if (p < 0)
        throw std::invalid_argument("Negative momentum given as argument");
      
      const double pz = p * cos(theta);
      const double pt = p * sin(theta);
      if (pt < 0)
        throw std::invalid_argument("Negative transverse momentum in calculation");
      const double px = pt * cos(phi);
      const double py = pt * sin(phi);
      //setPM(px, py, pz, mass);
      setP(px, py, pz);
      return *this;
    }

    /// Set the vector state from (theta,phi,pT) coordinates and the mass
    ///
    /// p = pt / sin(theta)
    /// pz = p cos(theta)
    /// E = sqrt(p^2 + mass^2)
    P3& setThetaPhiPt(double theta, double phi,  double pt) {
      if (theta < 0 || theta > M_PI)
        throw std::invalid_argument("Polar angle outside 0..pi given as argument");
      if (pt < 0)
        throw std::invalid_argument("Negative transverse momentum given as argument");
      const double p = pt / sin(theta);
      const double px = pt * cos(phi);
      const double py = pt * sin(phi);
      const double pz = p * cos(theta);
      
      setP(px, py, pz);
      return *this;
    }

   


    /// @name Coordinate getters
    //@{

    /// Get px^2
    double px2() const { return sqr(_x); }
    /// Get px
    double px() const { return _x; }
    /// Get py^2
    double py2() const { return sqr(_y); }
    /// Get py
    double py() const { return _y; }
    /// Get pz^2
    double pz2() const { return sqr(_z); }
    /// Get pz
    double pz() const { return _z; }
    
    double X() const { return _x; }
    /// Get py^2
    double Y() const { return _y; }
    /// Get pz^2
    double Z() const { return _z; }

    

    
    double p2() const { return sqr(px()) + sqr(py()) + sqr(pz()); }
    /// Get the spatial 3-vector |p|
    double p() const { return sqrt(p2()); }
    /// Get the spatial 3-vector |px^2 + py^2|
      double rho2() const { return sqr(px()) + sqr(py()); }
    double Perp2() const { return sqr(px()) + sqr(py()); }
    /// Get the spatial 3-vector sqrt|px^2 + py^2|
    double rho() const { return sqrt(rho2()); }
    double Perp() const { return sqrt(Perp2()); }
    /// Get the transverse momentum squared (same as rho2)
    double pT2() const { return Perp2(); }
    /// Get the transverse momentum (same as rho)
    double pT() const { return Perp(); }

    

    /// Get the spatial phi
    double phi() const { if (Perp2() == 0) return 0; else return atan2(py(),px()); }
    /// Get the spatial theta
    double theta() const { if (p2() == 0) return 0; else if (pz() == 0) return M_PI; else return atan2(rho(),pz()); }
    /// Get the spatial vector pseudorapidity
    double eta() const { return -log(tan( 0.5 * theta() )); } //< Optimise with a trig reln on tan(x/2) to avoid tan(atan(..)/2)?
    /// Get the spatial vector absolute pseudorapidity
    double abseta() const { return fabs(eta()); }

   

    //@}


    /// @name Calculations w.r.t. other P4 vectors
    //@{

    /// Spatial dot product
    double dot3(const P3& v) const { return px()*v.px() + py()*v.py() + pz()*v.pz(); }
    
    double Dot(const P3& v) const { return px()*v.px() + py()*v.py() + pz()*v.pz(); }

    /// Spatial angle to another P4 vector
    double angleTo(const P3& v) const { return acos( dot3(v) /p()/v.p() ); }
    double Angle(const P3& v) const { return acos( dot3(v) /p()/v.p() ); }  // for compatibility with TVector3
    /// Difference in phi between two vectors
    double deltaPhi(const P3& v) const { return deltaphi(phi(), v.phi()); }
    /// Difference in pseudorapidity between two vectors
    double deltaEta(const P3& v) const { return fabs(eta() - v.eta()); }
    /// Difference in rapidity between two vectors
    double deltaR2_eta(const P3& v) const { return sqr(deltaEta(v)) + sqr(deltaPhi(v)); }
    /// Difference in pseudorapidity-based R between two vectors
    double deltaR_eta(const P3& v) const { return sqrt(deltaR2_eta(v)); }
    /// Difference in rapidity-based R^2 between two vectors

    //hahahahaha
    P3 Cross(const P3& v) const { return P3(py()*v.pz()- pz()*v.py(),pz()*v.px()-px()-v.pz(),px()*v.py()-py()*v.px());} 

    // Reinvent the wheel for similarity with root TVector3
    // Need: Unit(), Mag(), 
    // https://root.cern.ch/doc/master/classTVector3.html

    double Mag() const { return p();}
    double Mag2() const { return p2();}
    double Theta() const { return theta();}
    double Phi() const { return phi();}

    double CosTheta() const { double pp = p(); if (pp == 0.0) return 0.0; else return pz()/pp; }

    // Unit vector parallel
    P3 Unit() const { 
      double pp=p();
      return P3(px()/pp,py()/pp,pz()/pp);
    }

    // Unit vector orthogonal ??
    
    //P3 Orthogonal() const {
    //}
    
    // rotate around a given axis, modifies the object
       // https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    // v = v|| + v_perp
    // v|| = (v.axis.Unit()) axis.Unit()
    // v_perp = v - v||


    // v * costheta + (k x v) sintheta + k (k.v) (1-costheta)

    P3& Rotate(const double &angle, const P3& axis) {
      P3 unitaxis=axis.Unit();
      double costheta=cos(angle);
      double sintheta=sin(angle);
      P3 kcrossv = Cross(unitaxis);
      double axismultiplier=Dot(unitaxis)*(1.0 - costheta);
      setPx(px()*costheta+kcrossv.px()*sintheta+unitaxis.px()*axismultiplier);
      setPy(py()*costheta+kcrossv.py()*sintheta+unitaxis.py()*axismultiplier);
      setPz(pz()*costheta+kcrossv.pz()*sintheta+unitaxis.pz()*axismultiplier);
      
      return (*this);
    }

    //@}

    /// @name Self-modifying operators
    //@{
    //P4  operator - () const { P4 rtn; return rtn.setPM(-_x, -_y, -_z, _m); } //< Not self-modifying...
    //P4& operator += (const P4& v) { double e = E() + v.E(); _x += v.px(); _y += v.py(); _z += v.pz(); _m = sqrt( sqr(e) - p2() ); return *this; }
    //P4& operator -= (const P4& v) { double e = E() - v.E(); _x -= v.px(); _y -= v.py(); _z -= v.pz(); _m = sqrt( sqr(e) - p2() ); return *this; }
    //P4& operator *= (double a) { _x *= a; _y *= a; _z *= a; _m *= a; return *this; }
    //P4& operator /= (double a) { _x /= a; _y /= a; _z /= a; _m /= a; return *this; }

    P3  operator - () const { P3 rtn; return rtn.setP(-_x, -_y, -_z); } //< Not self-modifying...
    P3& operator += (const P3& v) { _x += v.px(); _y += v.py(); _z += v.pz() ; return *this; }
    P3& operator -= (const P3& v) { _x -= v.px(); _y -= v.py(); _z -= v.pz() ; return *this; }
    P3& operator *= (double a) { _x *= a; _y *= a; _z *= a; return *this; }
    P3& operator /= (double a) { _x /= a; _y /= a; _z /= a; return *this; }
    
    //@}

  };


  /// @name String representations
  //@{

  /// Make a string representation of the vector
  inline std::string to_str(const P3& p3) {
    std::stringstream ss;
    ss << "(" << p3.px() << ", " << p3.py() << ", " << p3.pz() << ")";
    return ss.str();
  }

  /// Write a string representation of the vector to the provided stream
  inline std::ostream& operator <<(std::ostream& ostr, const P3& p3) {
    ostr << to_str(p3);
    return ostr;
  }

  //@}


  /// Convenience "external" functions
  //@{

  /// Lorentz inner product between two vectors
  inline double Dot(const P3& a, const P3& b) {
    return a.Dot(b);
  }


  /// Angle between two vectors
  inline double angle(const P3& a, const P3& b) {
    return a.angleTo(b);
  }


  /// Difference in phi between two vectors
  inline double deltaPhi(const P3& a, const P3& b) {
    return a.deltaPhi(b);
  }


  /// Difference in pseudorapidity between two vectors
  inline double deltaEta(const P3& a, const P3& b) {
    return a.deltaEta(b);
  }



  /// Difference in pseudorapidity-based R^2 between two vectors
  inline double deltaR2_eta(const P3& a, const P3& b) {
    return a.deltaR2_eta(b);
  }
  /// Difference in pseudorapidity-based R between two vectors
  inline double deltaR_eta(const P3& a, const P3& b) {
    return a.deltaR_eta(b);
  }



  //@}


  /// @name Operators taking two vectors
  //@{
  inline P3 operator + (const P3& a, const P3& b) { P3 rtn = a; return rtn += b; }
  inline P3 operator - (const P3& a, const P3& b) { P3 rtn = a; return rtn -= b; }
  inline P3 operator * (const P3& a, double f) { P3 rtn = a; return rtn *= f; }
  inline P3 operator * (double f, const P3& a) { P3 rtn = a; return rtn *= f; }
  inline P3 operator / (const P3& a, double f) { P3 rtn = a; return rtn /= f; }
  //@}








}
