#pragma once

#include "heputil.h"
#include "math.h"
#include <random>
#include <functional>
#include <cfloat> // for DBL_MAX

/**
 * 
 *  Header file to link functions used for isolation. * 
 * 
 * 
 **/

/// Compute d0 based on a production vertex and momentum, assuming no B field
double d0Calc(const HEP::P4 &xv, const HEP::P4 &p);

/// Compute dz based on a production vertex and momentum, assuming no B field
double dzCalc(const HEP::P4 &xv, const HEP::P4 &p);


/// Compute dz*sin \theta (used in various analyses) based on production vertex and momentum, assuming no B field
double dz_sintheta(const HEP::P4 &xv, const HEP::P4 &p);


/// Compute scalar sum of pT of charged tracks (charged hadrons, leptons) in a cone of radius DR around a momentum p4, subtracting the pT of the object itself
double sumpTisolation(const HEP::P4 &p4, const HEP::Event* evt, const double &DR);

/// Compute scalar sum of ET of all objects in a cone of radius DR around a momentum p4, subtracting the ET of the object itself
double sumETIsolation(const HEP::P4 &p4, const HEP::Event* evt, const double &DR);

double totalEventE(const HEP::Event* evt);
double medianEventE(const HEP::Event* evt);

double sumCaloE(const HEP::P4 &p4, const HEP::Event* evt, const double &DR);
double coneAreaApprox(const HEP::P4 &p4, const double &DR);


/// Classify  electron as "Gradient" using efficiency function, 1908.00005 page 37
bool IsGradientElectron_ATLAS_Run2(const  HEP::Particle* Lep, const HEP::Event* event, std::mt19937& engine);

/// Boolean to classify as loose/tight according to actual event information (not stochastically!) based on sumPTisolation and sumETisolation.  
/// To function therefore requires that we retain all tracks/hadrons in the events (so not with minimal event storage)
/// User provides the Delta R for pT and ET, plus the reference ratios of the pT and ET to the totals.
/// Used by the functions IsLooseElectronATLAS_Run2 etc.
bool IsLooseTightLepton(const  HEP::Particle* Lep, const HEP::Event* event,
		   const double DeltaRp, const double DeltaRE, const double pratio, const double Eratio);

/// @brief Check if a lepton is classed as highpTCaloOnly
/// @param Lep 
/// @param event 
/// @param DeltaRE 
/// @return Returns true if the sumETIsolation for DeltaRE is less than ETratio, defined as maximum of 1.5% of pT or 3.5 GeV.
bool IsHighPTCaloOnly(const HEP::Particle* Lep, const HEP::Event* event, const double DeltaRE);


/// mT of a vector with pTmiss
double mTcalc_simple( const HEP::P4 &v1, const HEP::P4 &v2);


bool ATLASradiusCalcLepton(const HEP::Particle* lepton,const HEP::Jet* jet);


/// @brief  Template function to filter a vector of pointers to objects by minimum pT and maximum abs(eta), returns the filtered vector
/// @tparam T1 
/// @param vec_t1 
/// @param pTmin 
/// @param absEtaMax 
/// @return 
template<typename T1> std::vector<T1*> filterObjects(std::vector<T1*> &vec_t1,const double &pTmin, const double &absEtaMax)
{
	// filters the particles/jets in a vector by pT and eta.
  std::vector<T1*> result;
  
  for(auto t1 : vec_t1)
  {
    if( (t1->pT() > pTmin) && (t1->abseta() < absEtaMax))
    {
      result.push_back(t1);
    }

  }

  return result;
  
}

/// @brief  Template function to filter a vector of pointers to objects by min pT, max abs(eta) AND a boolean condition function to be provided by the user, which may depend include a random element, provided by the random engine.
/// @tparam T1 
/// @param vec_t1 
/// @param pTmin 
/// @param absEtaMax 
/// @param evt 
/// @param engine 
/// @param select_condition 
/// @return 
template<typename T1> std::vector<T1*> filterObjects(std::vector<T1*> &vec_t1,const double &pTmin, const double &absEtaMax, const HEP::Event* evt,std::mt19937 &engine, bool (*select_condition)(T1*,const HEP::Event*, std::mt19937&))
{
	// filters the particles/jets in a vector by pT and eta.
  std::vector<T1*> result;
  
  for(auto t1 : vec_t1)
  {
    if( (t1->pT() > pTmin) && (t1->abseta() < absEtaMax))
    {
      //double efficiency=eff(t1,evt);
      //static std::uniform_real_distribution<double> rd(0.0,1.0);
      //if (rd(engine) < efficiency)
      if(select_condition(t1,evt,engine))
      {
        result.push_back(t1);
      }
    }

  }

  return result;
  
}

/// @brief Template function to filter a vector of pointers to objects only via a boolean condition function to be provided by the user, which may depend include a random element, provided by the random engine.
/// @tparam T1 
/// @param vec_t1 
/// @param evt 
/// @param engine 
/// @param select_condition 
/// @return 
template<typename T1> std::vector<T1*> filterObjects(std::vector<T1*> &vec_t1,const HEP::Event* evt,std::mt19937 &engine, bool (*select_condition)(T1*,const HEP::Event*, std::mt19937&) )
{
	// filters the particles/jets in a vector by pT and eta.
  std::vector<T1*> result;
  
  for(auto t1 : vec_t1)
  {
    
      if(select_condition(t1,evt,engine))
      {
        result.push_back(t1);
      }

  }

  return result;
  
}

/// @brief VOID Template function to filter a vector of pointers objects by minimum pT and max abs(eta), acts directly on the vector (so no need to have a return value). See also filterObjects which does the same thing but returns a filtered vector.
/// @tparam T1 
/// @param vec_t1 
/// @param pTmin 
/// @param absEtaMax 
template<typename T1> void filterPhaseSpace(std::vector<T1*> &vec_t1, const double &pTmin, const double &absEtaMax)
{
	// filters the particles/jets in a vector by pT and eta.

  
  auto it = vec_t1.begin();
  while(it != vec_t1.end())
  {
	  //T1* t1 = vec_t1[it];
	  //if( (t1->pT() > pTmin) || (t1->abseta() > absEtaMax))
    if( ((*it)->pT() < pTmin ) || ((*it)->abseta() > absEtaMax))
	  {
       it = vec_t1.erase(it);
        }
    else
      {
	it++;
      }
    
  }
  
}

/// equivalent to filterPhaseSpace, slightly different algorithm
template<typename T1> void delete_and_filterPhaseSpace(std::vector<T1*> &vec_t1, const double &pTmin, const double &absEtaMax)
{
	// filters the particles/jets in a vector by pT and eta.

  std::vector<T1*> out_vector;
  for(auto t1 : vec_t1)
  {
    if( (t1->pT() < pTmin ) || (t1->abseta() > absEtaMax))
	  {
       delete t1;
        }
    else
      {
        out_vector.push_back(t1);
      }

  }
  vec_t1=out_vector;
    
}





// Overlap Removal
//template<typename T1, typename T2> std::vector<T1*> Removal(std::vector<T1*> &v1, std::vector<T2*> &v2,const double &drmin);

// Overlap Removal for electrons with Delta R < min(0.4, 0.04 + 10 GeV/pT) around Jet
//template<typename T1, typename T2> std::vector<T1*> RemovalJE(std::vector<T1*> &v1, std::vector<T2*> &v2);


// Overlap Removal


/// template function for overlap removal. Takes two vectors of (pointers to) objects and removes those in the first that are within drmin of any object in the second; returns the filtered first vector. The objects must have the mom() method to give the HEP::P4 momentum. Removed objects are actually deleted.
template<typename T1, typename T2> std::vector<T1*> FullRemoval(std::vector<T1*> &v1, std::vector<T2*> &v2, const double &drmin)
{
  // Determining with objects should be removed, and free memory
  
  std::vector<bool> mask(v1.size(),false);

  for (unsigned int j=0;j<v1.size();j++)
  {
    for (unsigned int i=0; !mask[j] && i<v2.size();i++)
        {
            if (v2[i]->mom().deltaR_eta(v1[j]->mom()) < drmin) 
	            {
	                mask[j]=true;
	                //break;
	            }
        }
    };
  // Building the cleaned container
  std::vector<T1*> cleaned_v1;
  for (unsigned int i=0;i<v1.size();i++)
  {
    if (!mask[i]) { 
      cleaned_v1.push_back(v1[i]);
      }
      else
      {
        delete v1[i];
      }

  }

  return cleaned_v1;
};



/// template function for overlap removal. Takes two vectors of (pointers to) objects and removes those in the first that are within drmin of any object in the second; returns the filtered first vector. The objects must have the mom() method to give the HEP::P4 momentum.
template<typename T1, typename T2> std::vector<T1*>
  Removal(std::vector<T1*> &v1, std::vector<T2*> &v2,
  const double &drmin)
{
  // Determining with objects should be removed
  
  std::vector<bool> mask(v1.size(),false);

  for (unsigned int j=0;j<v1.size();j++)
  {
    for (unsigned int i=0; !mask[j] && i<v2.size();i++)
        {
            if (v2[i]->mom().deltaR_eta(v1[j]->mom()) < drmin) 
	            {
	                mask[j]=true;
	                //break;
	            }
        }
    };
  // Building the cleaned container
  std::vector<T1*> cleaned_v1;
  for (unsigned int i=0;i<v1.size();i++)
    if (!mask[i]) cleaned_v1.push_back(v1[i]);

  return cleaned_v1;
};


/// @brief template function for overlap removal. Takes two vectors of (pointers to) objects and removes those in the first that are within drmin of any object in the second AND meet a boolean condition specified by the user. 
/// @returns Returns the filtered first vector. The objects must have the mom() method to give the HEP::P4 momentum.
template<typename T1, typename T2> std::vector<T1*>
  Removal(std::vector<T1*> &v1, std::vector<T2*> &v2,
  const double &drmin,bool (*special_condition)(T1*,T2*))
{
  // Determining with objects should be removed
  
  std::vector<bool> mask(v1.size(),false);

  for (unsigned int j=0;j<v1.size();j++)
  {
    for (unsigned int i=0; !mask[j] && i<v2.size();i++)
        {
            if (special_condition(v1[j],v2[i]) && (v2[i]->mom().deltaR_eta(v1[j]->mom()) < drmin))
	            {
	                mask[j]=true;
	                //break;
	            }
        }
    };
  // Building the cleaned container
  std::vector<T1*> cleaned_v1;
  for (unsigned int i=0;i<v1.size();i++)
    if (!mask[i]) cleaned_v1.push_back(v1[i]);

  return cleaned_v1;
};




/// @brief template function for overlap removal. Takes one vector of (pointers to) objects. If two objects are within drmin of each other the one with the lowest pT is removed.
/// @returns Returns the filtered vector. The objects must have the mom() method to give the HEP::P4 momentum.
template<typename T1> std::vector<T1*> SelfRemoval(std::vector<T1*> &v1, const double &drmin)
{
  // Determining with objects should be removed -- keep the higher pT version. This only applies for electrons anyway
  
  std::vector<bool> mask(v1.size(),false);

      double tdr;
      for (unsigned int j=0;j<v1.size();j++){
	      for (unsigned int i=j+1;i<v1.size();i++) {

	        tdr=v1[i]->mom().deltaR_eta(v1[j]->mom());
	        if ((tdr < drmin)) 
		      {
           if(v1[i]->mom().pT() < v1[j]->mom().pT())
            {
		          mask[i]=true;
             }
            else
            {
              mask[j]=true;
            }
		      }

	      }
      };
   

  // Building the cleaned container
  std::vector<T1*> cleaned_v1;
  for (unsigned int i=0;i<v1.size();i++)
    if (!mask[i]) cleaned_v1.push_back(v1[i]);

  return cleaned_v1;
};



/// Overlap Removal for electrons with Delta R < min(0.4, 0.04 + 10 GeV/pT) around Jet
template<typename T1, typename T2> std::vector<T1*>
  RemovalJE(std::vector<T1*> &v1, std::vector<T2*> &v2)
{
  // Determining with objects should be removed
  double EPT,drmin;
  std::vector<bool> mask(v1.size(),false);
  for (unsigned int j=0;j<v1.size();j++)
    {
      EPT=v1[j]->pT();
      drmin=0.04+10.0/EPT;
      if(drmin > 0.4) drmin=0.4;
    
      for (unsigned int i=0;!mask[j] && i<v2.size();i++)
	      if (v2[i]->mom().deltaR_eta(v1[j]->mom()) < drmin)
	    {
        
	        mask[j]=true;
	        //break;
	    }
    }
  // Building the cleaned container
  std::vector<T1*> cleaned_v1;
  for (unsigned int i=0;i<v1.size();i++)
    if (!mask[i]) cleaned_v1.push_back(v1[i]);

  return cleaned_v1;
}



/*
Selection functions for ATLAS 
*/  

/// @brief Overlap removal 
/// @param Lep 
/// @param event 
/// @param engine 
/// @return 
bool IsLooseElectron_ATLAS_Run2(const  HEP::Particle* Lep, const HEP::Event* event, std::mt19937& engine);
bool IsTightElectron_ATLAS_Run2(const  HEP::Particle* Lep, const HEP::Event* event, std::mt19937& engine);

bool IsLooseTightMuon_ATLAS_Run2(const  HEP::Particle* Lep, const HEP::Event* event, std::mt19937& engine);

bool ED0Sigma5_ATLAS_Run2(const  HEP::Particle* Lep, const HEP::Event* event, std::mt19937& engine);
bool MuD0Sigma3_ATLAS_Run2(const  HEP::Particle* Lep, const HEP::Event* event, std::mt19937& engine);

bool BTag70DL1r_ATLAS_Run2(const  HEP::Jet* jet, const HEP::Event* event, std::mt19937& engine);


