/**
@file
My implementation of an analysis container. Written from scratch.
Functions similarly to MadAnalysis analyses, so we have to define
void init(), void Execute(std::mt19937 &engine), void Finalise() for a 
new analysis *but* I also require a constructor to be defined (with some standard stuff).
 

Signal regions defined via:

AddRegionSelection("region name");
   or 
AddRegionSelection(<vector of names>);

Cuts are defined via

AddCut("cut name",<signal region>);
   or 
AddCut("cut name",<vector of region names>);

Cuts are applied by
 
   ApplyCut(bool condition,const std::string& cutname);

Histogramming is done here via Yoda, so AddYodaHist1D in the init, FillYodaHist1D in the analysis
etc.


**/



#include "include/BaseAnalysis.h"


cut::cut()
  {
    weight_left=0.0;
    weight_square_left=0.0;
    number_left=0;
    name="";
    weights_left.clear();
    weights_squared_left.clear();
    
  };

 cut::cut(std::string cutname)
  {
    weight_left=0.0;
    weight_square_left=0.0;
    number_left=0;
    name=cutname;
    weights_left.clear();
    weights_squared_left.clear();
    
  }



cut::~cut() { };

cutflow::~cutflow() { for(auto cut : cuts) { delete cut; }; };

cutflow::cutflow() {
        _name = "" ;
         passed=true;
         cut* newcut = new cut("Initial Events");
         cuts.push_back(newcut);
    };
cutflow::cutflow(const std::string& name)
    {
        _name=name;
        passed=true;
        cut* newcut = new cut("Initial Events");
         cuts.push_back(newcut);
    };

void cutflow::addcut(const std::string& cutname)
    {
        cut* newcut = new cut(cutname);
        cutmap[cutname]=cuts.size(); // do this before pushing back ...
        cuts.push_back(newcut);
        
    };

void cutflow::applycut(const std::string& cutname,bool &condition, double &weight)
    {
        if(!this->passed) // "this" not strictly necessary. 
        {
            return;
        }
        int cutindex=cutmap[cutname];
        if (condition)
        {
            cuts[cutindex]->weight_left+=weight;
            cuts[cutindex]->weight_square_left+=pow(weight,2);
            cuts[cutindex]->number_left+=1;
        }
        else
        {
            this->passed=false;
        }

    };

void cutflow::applycut(const std::string& cutname,bool &condition, double &centralweight, vector<double> &weights)
    {
        if(!this->passed) // "this" not strictly necessary. 
        {
            return;
        }
        int cutindex=cutmap[cutname];
        if (condition)
        {

            cuts[cutindex]->weight_left+=centralweight;
            //cout << "centralweight " << centralweight << endl;
            //cuts[cutindex]->weight_square_left+=pow(centralweight,2);
            cuts[cutindex]->weight_square_left+=centralweight*centralweight;
            //cout << "square weight " << cuts[cutindex]->weight_square_left << endl;
            cuts[cutindex]->number_left+=1;
            int nweights=weights.size();
            int nweights_in_cut=cuts[cutindex]->weights_left.size();
            if(nweights_in_cut == 0)  // true for initial events, so initialise
            {
                for (int w=0; w<nweights; w++)
                {
                    cuts[cutindex]->weights_left.push_back(weights[w]);
                    cuts[cutindex]->weights_squared_left.push_back(weights[w]*weights[w]);
                }
            }
            else
            {
                nweights=std::min(nweights_in_cut,nweights);// can only add as many weights as we have
                for (int w=0; w<nweights; w++)
                {
                    cuts[cutindex]->weights_left[w]+=weights[w];
                    cuts[cutindex]->weights_squared_left[w]+=weights[w]*weights[w];
                }
            }
            //cuts[cutindex]->number_left+=1;
        }
        else
        {
            this->passed=false;
        }

    };



void cutflow::print(ostream &ss)
    {
        double initialweight=cuts[0]->weight_left;
        double eff=0.0;
        double Nevents=(double) cuts[0]->number_left;
        
        ss.precision(6);
        
        for(auto cut: cuts)
        {
            eff=(cut->weight_left)/initialweight;
            double reluncert=0.0;
            double uncert=0.0;
            if((Nevents > 0.0) && (eff > 0.0) && (eff <= 1.0))
            {
                uncert=sqrt(eff*(1.0-eff)/Nevents )*100.0;
                reluncert=uncert/eff;
            }
            
            ss << cut->name << " : " << cut->number_left << ", " << cut->weight_left << " -> efficiency: " << eff << " +/- " << uncert <<"%" << endl;

        }

    };

void cutflow::print(ostream &ss,vector<vector<int>> &weight_ensembles)
{
    ss.precision(6);
    double eff,staterr,syserrplus, syserrminus;
        for(auto cut: cuts)
        {
            geteff_and_all_errs(cut->name,weight_ensembles,eff,staterr,syserrplus,syserrminus);

            if((syserrplus < 1e-16) && (syserrminus < 1e-16))
            {
                ss  << cut->name << " : " << cut->number_left << ", " << cut->weight_left << " -> efficiency: " << std::scientific <<  std::setprecision(5) << eff  <<std::defaultfloat << std::setprecision(3) << " +/- " << staterr*100.0 << "% (stat)" << " +/- 0% (syst) " << endl;
            }
            else
            {
            ss  << cut->name << " : " << cut->number_left << ", " << cut->weight_left << " -> efficiency: "   << std::scientific << std::setprecision(5) << eff  <<std::defaultfloat << std::setprecision(3) << " +/- " << staterr*100.0 << "% (stat)" << " + " << syserrplus*100.0 << "% - " << syserrminus*100.0 << "% (syst) " << endl;
            }
        }

}


void cutflow::print()
    {
        print(std::cout);
    }

void cutflow::geteff_and_err(double &eff,double &err)
{
    geteff_and_err(this->cuts.back()->name,eff,err);
}
void cutflow::geteff_and_err(const std::string& cutname,double &eff,double &err)
{
    int cutindex=cutmap[cutname];
    double initialweight=cuts[0]->weight_left;
    int goodevents=cuts[0]->number_left;
    double finalweight=cuts[cutindex]->weight_left;
    eff=0.0;    
    if(initialweight>0.0) eff=finalweight/initialweight;
    double uncert=0.0;
    err=1.0; // default value should be order 1 uncertainty, basically if the efficiency is zero.
    if((goodevents > 0) && (eff > 0.0) && (eff < 1.0))
        {
            uncert=sqrt(eff*(1.0-eff)/goodevents);
            err=uncert/eff;
        } 
}

void cutflow::geteff_and_all_errs(const std::string& cutname,vector<vector<int>> &weight_ensembles,double &eff,double &staterr,double &syserrplus, double &syserrminus)
{
    int cutindex=cutmap[cutname];
    
    //double initialweight=cuts[0]->weight_left;
    //double finalweight=cuts[cutindex]->weight_left;
    double initialweight=cuts[0]->weight_left;
    double finalweight=cuts[cutindex]->weight_left;

    staterr=1.0; // default error of 100%
    syserrplus=0.0;
    syserrminus=0.0;
    eff=0.0;    
    if(initialweight>0.0) eff=finalweight/initialweight;

    // no relative error possible if efficiency is zero or too small

    if(fabs(eff) < 1e-16) return;
    
    double errplussq=0.0;
    double errminussq=0.0;
    int ensembleno=0;

    // debug code
    //cout << "Processing cutflow " << this->_name << endl;

    /*
    for(auto ensemble: weight_ensembles)
    {
        cout << "will check ensemble " << ensembleno << endl;
        ensembleno++;
        for (auto weightno : ensemble)
        {
            cout << "... with weight " << weightno << endl;
        }
    }
    */
    ensembleno=0;

    for(auto ensemble: weight_ensembles)
    {
        //cout << "Adding ensemble " << ensembleno << endl;
        ensembleno++;
        double thiseffplus=eff;
        double thiseffminus=eff;

        // get the envelope for each ensemble
        for (auto weightno : ensemble)
        {
            //cout << "Adding weight " << weightno << endl;
            if(weightno >= cuts[0]->weights_left.size() )
            {
                //cout << "greater than length of weights for this " << endl;
                continue;
            }
            /*cout << "Initial weights: " << cuts[0]->weights_left.size();
            for(auto ww : cuts[0]->weights_left)
            {   
                cout << ww << ", ";
            }
            cout << endl;*/

            /// For the first cut, we want uncertainty on the cross-section, not the efficiency
            double tinitialweight=cuts[0]->weights_left[weightno];
            if(cutindex==0)
            {
                tinitialweight=initialweight; 
            }            
            
            if(weightno >= cuts[cutindex]->weights_left.size() )
            {
               // cout << "greater than length of cut, no statistics possible " << endl;
                continue;
            }
            double tfinalweight=cuts[cutindex]->weights_left[weightno];
            double teff=0.0;

             if(tinitialweight>0.0) 
             { 
                teff=tfinalweight/tinitialweight; 
             }
             else continue;

             

             if (teff > thiseffplus) thiseffplus = teff;
             if (teff < thiseffminus) thiseffminus = teff;

        }
        // add errors of each group in quadrature
        errplussq +=pow(thiseffplus-eff,2);
        errminussq +=pow(eff-thiseffminus,2);

    }
    syserrplus=sqrt(errplussq);
    syserrminus=sqrt(errminussq);
    if(syserrplus < 1e-16) syserrplus = 0.0;
    if(syserrminus < 1e-16) syserrminus= 0.0;

    syserrplus=syserrplus/eff;
    syserrminus=syserrminus/eff;

    // now do stat error

    //double Neff=cuts[0]->weights_squared_left[0]/pow(initialweight,2);
    double Neff=pow(initialweight,2)/(cuts[0]->weight_square_left);
    double neff=0.0;
    if(cuts[cutindex]->weight_square_left > 0.0)
    {
        neff=pow(cuts[cutindex]->weight_left,2)/(cuts[cutindex]->weight_square_left);
    }
    //cout << "Doing stats: " << cuts[0]->weight_square_left << ", " << initialweight << ", " << eff << endl;
    staterr=0.0;
   
    if((neff > 0) && (eff > 0.0) && (eff < 1.0))
        {
            //staterr=eff*sqrt((1.0-eff)/neff);
            staterr=sqrt((1.0-eff)/neff); // return relative error
            if(staterr < 1e-16) staterr=0.0;
        } 
    /*
     double uncert=0.0;
    if((Neff > 0) && (eff > 0.0) && (eff < 1.0))
        {
            uncert=sqrt(eff*(1.0-eff)/Neff);
            //staterr=uncert/eff;
            staterr=uncert;
        } 
    */

}


bool BaseAnalysis::CheckConvergence(double margin)
{
    //bool isConverged=true;
    double eff, err;
    for(auto region : regions)
    {
        region.second->geteff_and_err(eff,err);
        if(err > margin)
        {
            return false;
        }
    }
    return true;
}

void BaseAnalysis::cleanup()
{
    
    for(auto cf : cutflows)
    {
       // cf->~cutflow();
        delete cf; // calls destructor, then deallocates
    }
    
    

   for(auto histo : _YodaHisto1Ds)
   {
       delete histo.second;
   }

   for(auto histo : _YodaHisto2Ds)
   {
       delete histo.second;
   }
   for(auto histo : _YodaProfile1Ds)
   {
       delete histo.second;
   }

}

void BaseAnalysis::setup()
{
    _totalevents = 0;
    _xsection = 0.0;
    _totalweight = 0.0;
    eventweight=0.0;
    cutflows.clear();
    regions.clear();
    regionmap.clear();
    analysisname="";
    //PythiaFillFunction="DEFAULT";
    DetectorFunction="DEFAULT";
    //this->_isMaster=isMaster;
    //_isMaster=true; // by default make it a "master" analysis, can optionally set as a "child"
}


void BaseAnalysis::set_xsection(double xs)
{
    _xsection=xs;
}


void BaseAnalysis::write_json(ostream &ss)
{
    // assume that the file already has opening brackets or whatever
    std::string sindent = "    ";

    ss << sindent << "\"" << analysisname <<  "\" : " << std::endl << sindent << sindent << "{" << std::endl;
    ss << sindent << sindent << "\"xs\" : " << this->_xsection << ", " << std::endl;
    ss << sindent << sindent << "\"nevents\" : " << _totalevents << ", " << std::endl;
    ss << sindent << sindent << "\"Regions\" : {" << std::endl;

     bool has_systematics = (this->_weight_ensembles.size() > 0);
        bool firstregion = true;
     for(auto region : regions )
    {   
        if (!firstregion) ss << "," << std::endl;
        firstregion = false;
         ss << sindent << sindent << sindent << "\"" << region.first << "\"" << ": {" << std::endl;
         cutflow* thiscutflow=region.second;

         ss << sindent << sindent << sindent << sindent << " \"Cuts\" : [" << std::endl;
         double eff,staterr,syserrplus, syserrminus,finalweight,finalweightsquare;
         bool firstcut=true;
            for(auto cut: thiscutflow->cuts)
            {
                if (!firstcut) ss << "," << std::endl;
                firstcut=false;
                 ss << sindent << sindent << sindent << sindent << sindent << "{" << std::endl;
                 ss << sindent << sindent << sindent << sindent << sindent << "\"name\" : \"" << cut->name << "\"," << std::endl;
                if(has_systematics) thiscutflow->geteff_and_all_errs(cut->name,this->_weight_ensembles,eff,staterr,syserrplus,syserrminus);
                else thiscutflow->geteff_and_err(cut->name,eff,staterr);

                //int cutindex=thiscutflow->cutmap[cut->name];
                //double finalweight=thiscutflow->cuts[cutindex]->weight_left;
                //double finalweightsquare=thiscutflow->cuts[cutindex]->weight_square_left;
                finalweight=cut->weight_left;
                finalweightsquare=cut->weight_square_left;
                ss << sindent << sindent << sindent << sindent << sindent << "\"efficiency\":" <<  std::scientific << std::setprecision(5) << eff  << "," << std::endl;
                ss << sindent << sindent << sindent  << sindent<< sindent << "\"events\":" <<  cut->number_left << "," << std::endl;
                ss << sindent << sindent << sindent  << sindent<< sindent << "\"stat\":" <<  std::defaultfloat << std::setprecision(3) << staterr*100.0  << "," << std::endl;
                ss << sindent << sindent << sindent  << sindent<< sindent << "\"weightsum\":" <<  std::scientific << std::setprecision(5) << finalweight  << "," << std::endl;
                ss << sindent << sindent << sindent  << sindent<< sindent << "\"sqweightsum\":" <<  std::scientific << std::setprecision(5) << finalweightsquare;
                if(has_systematics)
                {
                    ss << "," << std::endl;
                    ss << sindent << sindent << sindent << sindent << sindent << "\"sysup\":" <<  std::defaultfloat << std::setprecision(3) << syserrplus*100.0  << "," << std::endl;
                    ss << sindent << sindent << sindent  << sindent<< sindent << "\"sysdown\":" <<  std::defaultfloat << std::setprecision(3) << syserrminus*100.0 << std::endl;
                }

              

                ss << sindent << sindent << sindent  << sindent<< sindent << "}";

            }
            ss << std::endl << sindent << sindent << sindent << sindent << "]," << std::endl; // close cuts

              // now do final results for cutflow

                ss << sindent << sindent << sindent << sindent << "\"N0\":" << thiscutflow->cuts[0]->number_left << "," << std::endl;
                ss << sindent << sindent << sindent << sindent << "\"efficiency\":" <<  std::scientific << std::setprecision(5) << eff  << "," << std::endl;
                ss << sindent << sindent << sindent << sindent << "\"stat\":" <<  std::defaultfloat << std::setprecision(3) << staterr*100.0  << "," << std::endl;
                ss << sindent << sindent << sindent << sindent << "\"weightsum\":" <<  std::scientific << std::setprecision(5) << finalweight  << "," << std::endl;
                ss << sindent << sindent << sindent << sindent << "\"sqweightsum\":" <<  std::scientific << std::setprecision(5) << finalweightsquare;
                if(has_systematics)
                {
                    ss << "," << std::endl;
                    ss << sindent << sindent << sindent << sindent << "\"sysup\":" <<  std::defaultfloat << std::setprecision(3) << syserrplus*100.0  << "," << std::endl;
                    ss << sindent << sindent << sindent << sindent << "\"sysdown\":" <<  std::defaultfloat << std::setprecision(3) << syserrminus*100.0 << std::endl;
                }
            ss << sindent << sindent << sindent <<  "}" ;  // close the region
           
    }
    ss << std::endl << sindent << sindent  <<  "}" ; // close "Regions: {" 
    ss << std::endl << sindent << "}" ; // close analysis
}

/*
 ss.precision(6);
    double eff,staterr,syserrplus, syserrminus;
        for(auto cut: cuts)
        {
            geteff_and_all_errs(cut->name,weight_ensembles,eff,staterr,syserrplus,syserrminus);

            if((syserrplus < 1e-16) && (syserrminus < 1e-16))
            {
                ss  << cut->name << " : " << cut->number_left << ", " << cut->weight_left << " -> efficiency: " << std::scientific <<  std::setprecision(5) << eff  <<std::defaultfloat << std::setprecision(3) << " +/- " << staterr*100.0 << "% (stat)" << " +/- 0% (syst) " << endl;
            }
            else
            {
            ss  << cut->name << " : " << cut->number_left << ", " << cut->weight_left << " -> efficiency: "   << std::scientific << std::setprecision(5) << eff  <<std::defaultfloat << std::setprecision(3) << " +/- " << staterr*100.0 << "% (stat)" << " + " << syserrplus*100.0 << "% - " << syserrminus*100.0 << "% (syst) " << endl;
            }
        }


*/



void BaseAnalysis::write_results(ostream &ss)
{
    ss << "BLOCK PROCESS " << analysisname << std::endl;
    //ss << "XS  " << pythia.info.sigmaGen()*1.e12 << "   # (fb) " << std::endl;
    ss << "XS   " << this->_xsection <<  "   # (fb) " << std::endl;
    ss << "Events  " << _totalevents << std::endl;

    ss << "BLOCK EFFICIENCIES " << analysisname << " # name, eff, rel. uncert." << std::endl;

    bool has_systematics = (this->_weight_ensembles.size() > 0);

    for(auto region : regions )
    {
         cutflow* thiscutflow=region.second;

        double eff=0.0;
        
        //double staterr,syserrplus,syserrminus;
         //double reluncert=0.0;
        double staterr=1.0;
        if(has_systematics)
        {
            double syserrplus,syserrminus;
            thiscutflow->geteff_and_all_errs(thiscutflow->cuts.back()->name,this->_weight_ensembles,eff,staterr,syserrplus,syserrminus);
            ss << thiscutflow->_name << "   " << scientific <<  eff << "   " << staterr << "   " << syserrplus << "   " << syserrminus << std::endl;
        }
        else
        {
            thiscutflow->geteff_and_err(eff,staterr);
            ss << thiscutflow->_name << "   " << scientific <<  eff << "   " << staterr << std::endl;
        }


        
         /*
         double initialweight=thiscutflow->cuts[0]->weight_left;
         int goodevents=thiscutflow->cuts[0]->number_left;
         double finalweight=thiscutflow->cuts.back()->weight_left;
         double eff=0.0;
         if(initialweight>0.0) eff=finalweight/initialweight;

    
            double uncert=0.0;
            double reluncert=0.0;
        if((goodevents > 0) && (eff > 0.0) && (eff < 1.0))
        {
        uncert=sqrt(eff*(1.0-eff)/goodevents);
        reluncert=uncert/eff;
        }

        ss << thiscutflow->_name << "   " << scientific <<  eff << "   " << reluncert << std::endl;
        */
    }

}



bool  BaseAnalysis::ProcessEvent(HEP::Event *evnt){

        //eventweight=evnt->weights()[0];
        eventweight=evnt->weight(); // now gets central weight
        
        if(eventweight == 0.0)
        {
            return false;
        }
        


        _totalweight+=eventweight;
        _totalevents++;
        Event = evnt;

        //_totalweights=evnt->weights();
        
        
        /*
        weights_squared_left.clear();
        for(auto weight: weights_left)
        {
            weights_squared_left.push_back(weight*weight);
        }*/
        _eventweights.clear();
        for(auto weight : evnt->weights())
        {
           // cout << weight << ", ";
             _eventweights.push_back(weight);
        }
       //cout << endl;

        //cout << "Adding weights " << _eventweights.size() << endl;
        vector<double> squared_weights;
        //const vector<double>& weights 
        for(auto weight : _eventweights)
        {
            squared_weights.push_back(pow(weight,2));
        }

        for( auto region : regions)
        {
            region.second->passed=true; // second refers to second term in iterator
            region.second->cuts[0]->weight_left+=eventweight;
            region.second->cuts[0]->weight_square_left+=eventweight*eventweight;
            region.second->cuts[0]->number_left+=1;
            //cout << "Filling " << region.first << endl;
            //int nweights=evnt->weights().size();
            int nweights=_eventweights.size();
            if(nweights > 0) // we don't include the central weight in this list
            {
               // cout << "More than one weight!" << endl;
                if( region.second->cuts[0]->weights_left.size() ==0 )
                {
                        for(int w=0; w<nweights; w++)
                        {
                            region.second->cuts[0]->weights_left.push_back(_eventweights[w]);
                            region.second->cuts[0]->weights_squared_left.push_back(squared_weights[w]);
                        }
                           // region.second->cuts[0]->weights_left = _eventweights;
                            //region.second->cuts[0]->weights_squared_left = squared_weights;

                }
                else
                {
                    for (int w=0; w<nweights; w++)
                    {
                        region.second->cuts[0]->weights_left[w]+=_eventweights[w];
                        region.second->cuts[0]->weights_squared_left[w]+=squared_weights[w];
                    }
                }

            }
            //

           
            /*
            int nweights_in_cut=cuts[cutindex]->weights_left.size();
            if(nweights_in_cut == 0)  // true for initial events, so initialise
            {
                for (int w=0; w<nweights; w++)
                {
                    cuts[cutindex]->weights_left.push_back(weights[w]);
                    cuts[cutindex]->weights_squared_left.push_back(weights[w]*weights[w]);
                }
            }
            */


        }

        return true;

    };

void BaseAnalysis::AddRegionSelection(const std::string &region_name) {
        cutflow* newreg=new cutflow(region_name);
        this->cutflows.push_back(newreg);
        regions[region_name] = newreg;
    };


void BaseAnalysis::AddCut(const std::string& cutname,const std::vector<std::string> region_names) {
    
    for(auto name: region_names)
    {
        
        this->AddCut(cutname,name);

    }
};


void BaseAnalysis::AddCut(const std::string& cutname,const std::string& region_name)
    {
        regions[region_name]->addcut(cutname);
        regionmap[cutname].push_back(regions[region_name]); // pointer to the cutflow
    }

    

void BaseAnalysis::ApplyCut(bool condition, const std::string& cutname)
    {
        vector<cutflow*> region_names=regionmap[cutname];
        for ( cutflow* region : region_names)
        {
            //std::cout << "Applying cut with weight " << this->eventweight << std::endl;
            //region->applycut(cutname,condition,this->eventweight);
            if(this->_eventweights.size() > 0)
            {
                region->applycut(cutname,condition,this->eventweight,this->_eventweights);
            }
            else
            {
                region->applycut(cutname,condition,this->eventweight);
            }
        }

    }



void BaseAnalysis::print_cutflows()
{
    BaseAnalysis::print_cutflows(std::cout);

}

void BaseAnalysis::print_cutflows(ostream &os)
{

    
    for( auto region : regions)
    {
        os << "============================= " << region.first << " =============================" << endl;
        if(this->_weight_ensembles.size() > 0)
        {
            region.second->print(os,_weight_ensembles);
        }
        else
        {
            region.second->print(os);
        }
        
        
        
    }
    os << "==========================================================" << endl;
}

void BaseAnalysis::set_weight(double weight)
{
    this->eventweight = weight;

    // NO! Don't want to mess with the underlying event. 
    // We don't use it for the cutflows and
    // it may be used by other analyses
    //Event->set_weight(weight);


}

void BaseAnalysis::set_weights(vector<double> &weights)
{
    this->_eventweights = weights;
}


void BaseAnalysis::Reweight(double weight_multiplier)
{
    this->eventweight *= weight_multiplier;

    for(int w =0; w< this->_eventweights.size(); w++)
    {
        this->_eventweights[w] *= weight_multiplier;
    }


    // NO! Don't want to mess with the underlying event. 
    // We don't use it for the cutflows and
    // it may be used by other analyses
    //Event->set_weight(this->eventweight);


}

    
double BaseAnalysis::get_weight()
{
    return this->eventweight;
}


    void BaseAnalysis::add ( BaseAnalysis& B)
    {   
        this->_totalweight+=B._totalweight;
        this-> _totalevents+=B._totalevents;
        this->_weight_ensembles=B._weight_ensembles;
        for(auto region : this->regions )
        {
         cutflow* thiscutflow=region.second;
         
         
         cutflow* thatcutflow=B.regions[region.first];
         
            for(int cc=0; cc< thatcutflow->cuts.size(); cc++)
            {
                thiscutflow->cuts[cc]->weight_left+=thatcutflow->cuts[cc]->weight_left;
                thiscutflow->cuts[cc]->weight_square_left+=thatcutflow->cuts[cc]->weight_square_left;
                if(thiscutflow->cuts[cc]->weights_left.size() ==0)  // don't actually have any weights, probably master analysis
                {
                    for(int w =0;w < thatcutflow->cuts[cc]->weights_left.size(); w++)
                    {
                        thiscutflow->cuts[cc]->weights_left.push_back(thatcutflow->cuts[cc]->weights_left[w]);
                        thiscutflow->cuts[cc]->weights_squared_left.push_back(thatcutflow->cuts[cc]->weights_squared_left[w]);
                    }
                }
                else
                {
                    for(int w =0;w < thatcutflow->cuts[cc]->weights_left.size(); w++)
                    {
                        thiscutflow->cuts[cc]->weights_left[w]+=thatcutflow->cuts[cc]->weights_left[w];
                        thiscutflow->cuts[cc]->weights_squared_left[w]+=thatcutflow->cuts[cc]->weights_squared_left[w];
                    }
                }
                thiscutflow->cuts[cc]->number_left+=thatcutflow->cuts[cc]->number_left;
            }
            
        
        }

        for(auto histomap : _YodaHisto1Ds)
        {
            *(histomap.second) += *(B.GetYodaHisto1D(histomap.first));

        }

        for(auto histomap : _YodaHisto2Ds)
        {
            *(histomap.second) += *(B.GetYodaHisto2D(histomap.first));

        }

        for(auto histomap : _YodaProfile1Ds)
        {
            *(histomap.second) += *(B.GetYodaProfile1D(histomap.first));

        }
       

    }




//////////////////////////////////////////////////////////////////////
// YODA Stuff
//

void BaseAnalysis::AddYodaHisto1D(const std::string& objectname)
{
    // Adds an empty Histo1D. The user can then manipulate it at will using the 
    // GetYodaHisto1D method ... or use my handily written fill function.

    //static string nullpath="";
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    string path=this->analysisname+"/"+objectname;

    YODA::Histo1D* tHist = new YODA::Histo1D(path, objectname);
    _YodaHisto1Ds[objectname] = tHist;

}
void BaseAnalysis::AddYodaHisto1D(const std::string& objectname, size_t nbins, double lower, double upper)
{
    
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    string path=this->analysisname+"/"+objectname;
    YODA::Histo1D* tHist = new YODA::Histo1D(nbins, lower, upper, path, objectname);
    _YodaHisto1Ds[objectname] = tHist;

}

void BaseAnalysis::AddYodaHisto1D(const std::string& objectname, const std::vector<double>& binedges)
{
    
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    string path=this->analysisname+"/"+objectname;
    YODA::Histo1D* tHist = new YODA::Histo1D(binedges, path, objectname);
    _YodaHisto1Ds[objectname] = tHist;

}

void BaseAnalysis::FillYodaHisto1D(const std::string& objectname, double x, double fraction)
{
    _YodaHisto1Ds[objectname]->fill(x,eventweight,fraction);
}

YODA::Histo1D* BaseAnalysis::GetYodaHisto1D(const std::string& objectname)
{
    return _YodaHisto1Ds[objectname];
}

///////////////////////////// profile1d
void BaseAnalysis::AddYodaProfile1D(const std::string& objectname)
{
   
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    YODA::Profile1D* tHist = new YODA::Profile1D(objectname, objectname);
    _YodaProfile1Ds[objectname] = tHist;

}
void BaseAnalysis::AddYodaProfile1D(const std::string& objectname, size_t nbins, double lower, double upper)
{
    
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    YODA::Profile1D* tHist = new YODA::Profile1D(nbins, lower, upper, objectname, objectname);
    _YodaProfile1Ds[objectname] = tHist;

}

void BaseAnalysis::AddYodaProfile1D(const std::string& objectname, const std::vector<double>& binedges)
{
    
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    YODA::Profile1D* tHist = new YODA::Profile1D(binedges, objectname, objectname);
    _YodaProfile1Ds[objectname] = tHist;

}

void BaseAnalysis::FillYodaProfile1D(const std::string& objectname, double x, double y, double fraction)
{
    _YodaProfile1Ds[objectname]->fill(x,y,eventweight,fraction);
}

YODA::Profile1D* BaseAnalysis::GetYodaProfile1D(const std::string& objectname)
{
    return _YodaProfile1Ds[objectname];
}

///////////////////////////// end profile1d


///////////////////////////// histo2d
void BaseAnalysis::AddYodaHisto2D(const std::string& objectname)
{
    //static string nullpath="";
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    YODA::Histo2D* tHist = new YODA::Histo2D(objectname, objectname);
    _YodaHisto2Ds[objectname] = tHist;

}
void BaseAnalysis::AddYodaHisto2D(const std::string& objectname, size_t nbinsX,size_t nbinsY, double lowerX, double upperX, double lowerY, double upperY)
{
    //static string nullpath="";
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    YODA::Histo2D* tHist = new YODA::Histo2D(nbinsX,nbinsY,lowerX,upperX,lowerY,upperY, objectname, objectname);
    _YodaHisto2Ds[objectname] = tHist;

}

void BaseAnalysis::AddYodaHisto2D(const std::string& objectname, const std::vector<double>& xedges, const std::vector<double>& yedges)
{
    //static string nullpath="";
    // First copy of "objectname" here is the "path". I just set this to be the same as the object name, so that when
    // plotting it is produced in the same directory
    YODA::Histo2D* tHist = new YODA::Histo2D(xedges, yedges, objectname, objectname);
    _YodaHisto2Ds[objectname] = tHist;

}

void BaseAnalysis::FillYodaHisto2D(const std::string& objectname, double x, double y, double fraction)
{
    _YodaHisto2Ds[objectname]->fill(x,y,eventweight,fraction);
}

YODA::Histo2D* BaseAnalysis::GetYodaHisto2D(const std::string& objectname)
{
    return _YodaHisto2Ds[objectname];
}

///////////////////////////// end Histo2d




void BaseAnalysis::WriteHistos(ostream &os)
{
    for(auto histomap : _YodaHisto1Ds)
        {
            YODA::WriterYODA::write(os, *(histomap.second));

        }

    for(auto histomap : _YodaHisto2Ds)
        {
            YODA::WriterYODA::write(os, *(histomap.second));

        }

    for(auto histomap : _YodaProfile1Ds)
        {
            YODA::WriterYODA::write(os, *(histomap.second));

        }

}
