/**
 * @file running.cc
 * @brief Routines for handling initialisation and the progress bar.
 * 
 */
 


#include "include/running.h"


// So that we can define the analyses to include dynamically and choose which to include at runtime
    BaseAnalysis* mkAnalysis(const std::string name)
    {
      MAP_ANALYSES(IF_X_RTN_CREATE_ANA_X)
#ifdef ONNXPATH
      MAP_ONNX_ANALYSES(IF_X_RTN_CREATE_ANA_X)
#endif
		
	
      throw std::runtime_error("The analysis " + name + " is not a known analysis.");
      return nullptr;
    }

void CreateAnalysisList(vector<string> &analysislist)
{
    // This function adds *all* available analyses to the run 
    analysislist.clear();
    MAP_ANALYSES(APPEND_ANALYSIS_TO_LIST)
#ifdef ONNXPATH
    MAP_ONNX_ANALYSES(APPEND_ANALYSIS_TO_LIST)
#endif  
}



void progressbar(float progress, string message)
{
  static int barWidth = 50;
  int pos = barWidth * progress;
  std::string lineout;

 lineout += "[";
  for (int i = 0; i < barWidth; ++i) {
    if (i < pos) lineout += "\033[1;32m=\033[0m" ;  // makes the text blue see https://stackoverflow.com/questions/2616906/how-do-i-output-coloured-text-to-a-linux-terminal
        else if (i == pos) lineout += "\033[1;32m>\033[0m";
        else lineout += " ";
    }
  lineout += "] " + to_string(int(progress * 100.0)) + " % (" + message + ")";

  for(int i =lineout.length(); i<shellWidth; i++)
  {
    lineout += " ";
  }
  
  std::cout << lineout << "\r";

  std::cout.flush();
};


/////////////////////////////////////////////////
// Settings stuff




SettingsContainer::SettingsContainer()
{
    this->SetDefaults();
}

/*! Default settings are given here */
void SettingsContainer::SetDefaults()
{
    ncores=1;  //!< default to 1 core
    nevents=1000000; //!< million events should be enough as a default for most files
    xs=-1.0; //!< default cross-section negative -> will determine from pythia if running
    Rjet=0.4;  //!< Rjet set to 0.4 by default
    lhefile="";
    outfile="test.slha";
    cfgfile="test.cfg";
    histofile="test.yoda";
    cutflowfile="test.eff";
    hepmcfile="test.hepmc";
    jsonfile="hackanalysis.json";
    writehistos=true;
    includePileUp=false;
    npileup=29;
    pileupfilename="minbias.dat.gz";
    Analyses_To_Run = {};
    checkconvergence=false;
    convergencemargin=0.1;
    store_events=false;  //!< store event files in HackAnalysis event format ... default is false
    store_hadrons=false; //!< keep hadrons in the written event files ... default is false
}


namespace HASLHA{

// Routines here rely on code adapted from the Pythia SLHA reader

void toUpper(string& name, bool trim=true) {

  // stolen shamelessly from pythia, with modification
  // Copy string without initial and trailing blanks or escape characters.
  string temp = name;
  if (trim) {
    if (name.find_first_not_of(" \n\t\v\b\r\f\a") == string::npos) return;
    int firstChar = name.find_first_not_of(" \n\t\v\b\r\f\a");
    int lastChar  = name.find_last_not_of(" \n\t\v\b\r\f\a");
    temp          = name.substr( firstChar, lastChar + 1 - firstChar);
  }

  // Convert to Uppercase letter by letter.
  for (int i = 0; i < int(temp.length()); ++i) temp[i] = toupper(temp[i]);
  name = temp;

}

void ReadQNUMBERS(std::string &fname)
{
    ifstream is;

    is.open(fname, std::ifstream::in);
    string line="";
   bool readQblock=false;
 string blockName,blockIn;
 
 int pdg,Q3,colour;
 bool foundcolour=false;
 bool foundQ3=false;
 
  while ( getline(is, line) )
    {
      if (line.find("#") == 0) continue;
      if (line.size() == 0) continue;
      if (line.size() == 1 && line.substr(0,1) == " ") continue;

      if (line.find("#") != string::npos) {
	line.erase(line.find("#"),line.length()-line.find("#"));
      }
      toUpper(line);
      if (line.find("BLOCK") <= 1) {

	// new block: flush
	if(readQblock && foundcolour && foundQ3)
	      {
          cout<< "Adding particle " << pdg << " with 3*charge: " << Q3 << ", colour rep: " << colour <<endl; 
		PDG::AddParticle(pdg,Q3,colour);
	      }
	    foundcolour=false;
	    foundQ3=false;
	    
	blockIn=line ;
	if (blockIn.find("QNUMBERS") == string::npos)
	  {
	    readQblock=false;
	    continue;
	  }

	readQblock=true;
	foundcolour=false;
	foundQ3=false;
	//int nameBegin=6 ;
	//int nameEnd=blockIn.find(" ",7);
	//blockName=blockIn.substr(nameBegin,nameEnd-nameBegin);

	// should be BLOCK QNUMBERS pdg [antipdg]
	int pdgBegin=blockIn.find(" ",7)+1;
        int pdgEnd=blockIn.find(" ",pdgBegin);
        string pdgString = blockIn.substr(pdgBegin,pdgEnd-pdgBegin);
	pdg = stoi(pdgString);
	continue;
	
      }

      if(!readQblock) continue;

      // remove double spaces
      while (line.find("  ") != string::npos)
      	 {
	   int firstdoublespace = line.find_first_of("  ");
	   line.replace(firstdoublespace, 2, " ");
      	 }
      // remove tabs
      while (line.find("\t") != string::npos)
      	 {
	   int firsttab = line.find_first_of("\t");
	   line.replace(firsttab, 1, " ");
      	 }
      // read entry to QBlock
      // istream& getline (istream& is, string& str, char delim);
      std::istringstream sline(line);
      std::string entry;
      std::vector<int> data;
      while(std::getline(sline,entry, ' '))
      	{
	  try{
	    data.push_back(stoi(entry));
	  }
	  catch (int error) {
	    cout << "Unexpected data in QNUMBERS block: " << error << endl << line << endl;
	    break;
	  }
	}
      if(data.size() > 1)
	{
	  if(data[0] == 1)
	    {
	      Q3 = data[1];
	      foundQ3=true;
	    }
	  else if(data[0] == 3)
	    {
	      colour = data[1];
	      foundcolour=true;
	   }
	}
    }
  // flush any final block
  if(readQblock && foundcolour && foundQ3)
	      {
    cout<< "Adding particle " << pdg << " with 3*charge: " << Q3 << ", colour rep: " << colour <<endl;  
		PDG::AddParticle(pdg,Q3,colour);
	      }

        return;


}


}

/// ReadYAML options
void SettingsContainer::ReadYAML(std::string & YAMLfname)
{
    this->SetDefaults();
    YODA_YAML::Node config = YODA_YAML::LoadFile(YAMLfname);
    for (const auto& it : config["analyses"])
   {
       Analyses_To_Run.push_back(it.as<std::string>());
       cout << "Adding analysis " << it.as<std::string>() << endl;  
   }


  // Add particles in the YAML file under "new particles":
    for (const auto& it : config["new particles"])
   {
     int Q3=0;
        int colour=1;
        std::string spid = it.first.as<std::string>();
        int pid = stoi(spid);
YODA_YAML::Node tnode=config["new particles"][spid];
       
       if(tnode["3Q"]) Q3 = stoi(tnode["3Q"].as<std::string>());
       if(tnode["colour"]) colour = stoi(tnode["colour"].as<std::string>());   
       cout<< "Adding particle " << pid << " with 3*charge: " << Q3 << ", colour rep: " << colour <<endl;  
        PDG::AddParticle(pid,Q3,colour);
   }


    /// Add particles from an SLHA file with QNUMBERS with option "QNUMBERS file"
    YODA_YAML::Node yamlres = config["settings"]["QNUMBERS file"];
            if(yamlres)
            {
                std::string qnumbers;
                qnumbers=yamlres.as<std::string>();
                HASLHA::ReadQNUMBERS(qnumbers);
            }
  

  this->pythia_strings.clear();
   for(const auto& it: config["Pythia Extras"])
   {
     pythia_strings.push_back(it.as<std::string>());
   }
    /// (incomplete) list of options for the YAML file:
    this->SetValue(this->ncores,"cores",config);  ///  cores: number of cores
    this->SetValue(this->cfgfile,"Config file",config);  /// Config file: cfg file for pythia
    this->SetValue(this->lhefile,"LHE file",config);    /// LHE file -> .lhe or .lhe.gz file for analysePYTHIA_LHE. For multicore operation they should be numbered by thisname0.lhe.gz, thisname1.lhe.gz, ... up to ncores-1.
    this->SetValue(this->hepmcfile,"HEPMC file",config);  /// HEPMC file name
    this->SetValue(this->jsonfile,"JSON file",config);    /// Name for the output json file
    this->SetValue(this->xs,"Cross-section",config);  /// Option to supply a cross-section, either to override the pythia-determined value or for hepmc files
    this->SetValue(this->Rjet,"Rjet",config);   /// Rjet -> jet radius, defaults to 0.4
    this->SetValue(this->nevents,"nevents",config);   /// nevents -> number of events
    this->SetValue(this->includePileUp,"Include Pileup",config);
    this->SetValue(this->store_events,"Store Events",config);
    this->SetValue(this->store_hadrons,"Store Hadrons",config);
    this->SetValue(this->npileup,"Average Pileup",config);
    this->SetValue(this->checkconvergence,"Check Convergence",config);
    this->SetValue(this->convergencemargin,"Convergence Margin",config);

    this->SetValue(this->pileupfilename,"Pileup Filename",config);
    this->SetValue(this->writehistos,"Write Histograms",config);
    this->SetValue(this->outfile,"Efficiency Filename",config);
    this->SetValue(this->cutflowfile,"Cutflow Filename",config);
    this->SetValue(this->histofile,"Histogram Filename",config);



}
